import abc
import logging

import numpy as np
from joblib import Memory

from aspire.image import Image

logger = logging.getLogger(__name__)


class Xform(abc.ABC):
    """
    An Xform is anything that implements a `forward` method (and an `adjoint` method, in the case of a LinearXform),
    that takes in a square Image object and spits out a square Image object corresponding to forward/adjoint operations.

    It does this by setting up whatever internal data structures it needs to set up in its constructor.
    Xform objects usually set up data structures that are typically *larger* than the depth of the Image
    object that they expect to encounter during any invocation of `forward` or `adjoint`.

    At runtime, it gets an Image object (a thin wrapper on a n x L x L ndarray),
    as well as numeric `indices` (a numpy array of index values) that correspond to index values (the 'window') of
    the incoming Image object within the context of all images (i.e. an Image object of N x L x L, where N represents
    the number of total images that can ever pass through this Xform.

    At runtime, The Xform object may choose to ignore `indices` altogether (e.g. a Xform that downsamples all incoming
    images to a constant resolution won't care what `indices` they correspond to), or may do something with it (e.g. a
    Xform that shifts images by varying offsets depending on their indices).
    """

    class XformActiveContextManager:
        """
        This inner class allows us to temporarily enable/disable a Xform object, by tweaking its `active`
        attribute on enter/exit.
        """

        def __init__(self, xform, active):
            self.xform = xform
            self.active = active

        def __enter__(self):
            self.xform_old_state = self.xform.active
            self.xform.active = self.active

        def __exit__(self, exc_type, exc_value, exc_traceback):
            self.xform.active = self.xform_old_state

    def __init__(self, active=True):
        """
        Create a Xform object that works at a specific resolution.

        :param active: A boolean indicating whether the Xform is active. True by default.
        """
        self.active = active

    def forward(self, im, indices=None):
        """
        Apply forward transformation for this Xform object to an Image object.

        :param im: The incoming Image object of depth `n`, on which to apply the forward transformation.
        :param indices: The indices to use within this Xform. If unspecified, [0..n) is used.
        :return: An Image object after applying the forward transformation.
        """
        if not self.active:
            return im
        if indices is None:
            indices = np.arange(im.n_images)
        return self._forward(im, indices=indices)

    @abc.abstractmethod
    def _forward(self, im, indices):
        """
        Subclasses must implement the _forward method applicable to im/indices.
        """

    def enabled(self):
        """
        Enable this Xform in a context manager, regardless of its `active` attribute value.
        :return: A context manager in which this Xform is enabled.
        """
        return Xform.XformActiveContextManager(self, active=True)

    def disabled(self):
        """
        Disable this Xform in a context manager, regardless of its `active` attribute value.
        :return: A context manager in which this Xform is disabled.
        """
        return Xform.XformActiveContextManager(self, active=False)

    def __str__(self):
        return self.__class__.__name__


class LinearXform(Xform):
    """
    A LinearXform is a Xform that additionally provides an `adjoint` method, similar to the `forward` method.
    """

    def adjoint(self, im, indices=None):
        """
        Apply adjoint transformation for this Xform object to an Image object.

        :param im: The incoming Image object of depth `n`, on which to apply the adjoint transformation.
        :param indices: The indices to use within this Xform. If unspecified, [0..n) is used.
        :return: An Image object after applying the adjoint transformation.
        """
        if not self.active:
            return im
        if indices is None:
            indices = np.arange(im.n_images)
        return self._adjoint(im, indices=indices)

    @abc.abstractmethod
    def _adjoint(self, im, indices):
        """
        Subclasses must implement the _adjoint method applicable to im/indices.
        """


class SymmetricXform(LinearXform):
    """
    A Symmetric Xform is a LinearXform where the forward/adjoint operations are identical.
    """

    def _adjoint(self, im, indices=None):
        return self._forward(im, indices)


class Multiply(SymmetricXform):
    """
    A Xform that changes the amplitudes of a stack of 2D images (in the form of an Image object) by multiplying all
    pixels of a single 2D  image by a constant factor.
    """

    def __init__(self, factor):
        """
        Initialize a Multiply Xform using specified factors

        :param factor: A float/int or an ndarray of scalar factors to use for amplitude multiplication.
        """
        super().__init__()
        self.multipliers = np.array(factor)

    def _forward(self, im, indices):
        if self.multipliers.size == 1:  # if we have a scalar multiplier
            im_new = im * self.multipliers
        else:
            im_new = im * self.multipliers[indices]

        return im_new

    def __str__(self):
        if self.multipliers.size == 1:
            return f"Multiply ({self.multipliers})"
        return "Multiply (multiple)"


class Shift(LinearXform):
    """
    A Xform that shifts pixels of a stack of 2D images (in the form of an Image object)by offsetting all pixels of a
    single 2D image by constant x/y offsets.
    """

    def __init__(self, shifts):
        """
        Initialize a Shift Xform using a Numpy array of shift values.

        :param shifts: An ndarray of shape (2) or (n, 2)
        """
        super().__init__()
        self.shifts = np.array(shifts)

    def _forward(self, im, indices):
        if self.shifts.ndim == 1:
            im_new = im.shift(self.shifts)
        else:
            im_new = im.shift(self.shifts[indices])

        return im_new

    def _adjoint(self, im, indices):
        if self.shifts.ndim == 1:
            im_new = im.shift(-self.shifts)
        else:
            im_new = im.shift(-self.shifts[indices])

        return im_new

    def __str__(self):
        if self.shifts.ndim == 1:
            return f"Shift ({self.shifts})"
        return "Shift (multiple)"


class Downsample(LinearXform):
    """
    A Xform that downsamples an Image object to a resolution specified by this Xform's resolution.
    """

    def __init__(self, resolution):
        self.resolution = resolution
        super().__init__()

    def _forward(self, im, indices):
        return im.downsample(self.resolution)

    def _adjoint(self, im, indices):
        # TODO: Implement up-sampling with zero-padding
        raise NotImplementedError("Adjoint of downsampling not implemented yet.")

    def __str__(self):
        return f"Downsample (Resolution {self.resolution})"


class FilterXform(SymmetricXform):
    """
    A `Xform` that applies a single `Filter` object to a stack of 2D images (as an Image object).
    """

    def __init__(self, filter):
        """
        Initialize the Filter `Xform` using a `Filter` object

        :param filter: An object of type `aspire.operators.Filter`
        """
        super().__init__()
        self.filter = filter

    def _forward(self, im, indices):
        return im.filter(self.filter)

    def __str__(self):
        return f"FilterXform ({self.filter})"


class Add(Xform):
    """
    A Xform that add the density of a stack of 2D images (in the form of an Image object)
    by an offset value.
    """

    def __init__(self, addend):
        """
        Initialize an Add Xform using a Numpy array of predefined values.

        :param addend: An float/int or an ndarray of shape (n,)
        """
        super().__init__()
        self.addend = np.array(addend)

    def _forward(self, im, indices):
        if self.addend.size == 1:
            im_new = im + self.addend
        else:
            im_new = im + self.addend[indices]

        return im_new

    def __str__(self):
        if self.addend.size == 1:
            return f"Add ({self.addend})"
        return "Add (multiple)"


class LambdaXform(Xform):
    """
    A `Xform` that applies a predefined function to a stack of 2D images (as an Image object).
    """

    def __init__(self, lambda_fun, *args, **kwargs):
        """
        Initialize the `LambdaXform` using a lambda function

        :param lambda_fun: Predefine lambda function
        :param *args: Positional arguments
        :param **kwargs: Keyword arguments
        """
        super().__init__()
        self.lambda_fun = lambda_fun
        self.args = args
        self.kwargs = kwargs

    def _forward(self, im, indices):
        im_in = im.asnumpy()
        im_out = self.lambda_fun(im_in, *self.args, **self.kwargs)

        return Image(im_out)

    def __str__(self):
        return f"LambdaXform ({self.lambda_fun.__name__})"


class IndexedXform(Xform):
    """
    An IndexedXform is a Xform where individual Xform objects are used at specific indices of the incoming Image object.

    This extra layer of abstraction is used because individual Xform objects are typically capable of dealing with
    a stack of images (an Image object) in an efficient manner. The IndexedXform class groups the invocation of each
    of its unique Xform objects, so that calls to individual Xform objects within it is minimized, and equals the
    number of unique Xforms found.
    """

    def __init__(self, unique_xforms, indices=None):
        if indices is None:
            indices = np.arange(0, len(unique_xforms))
        else:
            # Ensure we're dealing with a numpy array since we'll be utilizing numpy indexing later on.
            indices = np.array(indices)
            assert np.min(indices) >= 0
            assert np.max(indices) < len(unique_xforms)

        super().__init__()

        self.n_indices = len(indices)
        self.indices = indices
        self.n_xforms = len(unique_xforms)
        self.unique_xforms = unique_xforms

        # A list of references to individual Xform objects, with possibly multiple references pointing to
        # the same Xform object.
        self.xforms = [unique_xforms[i] for i in indices]

    def _indexed_operation(self, im, indices, which):
        """
        Apply either a forward or adjoint transformations to `im`, depending on the value of the 'which' parameter.

        :param im: The incoming Image object on which to apply the forward or adjoint transformations.
        :param indices: The indices of the transformations to apply.
        :param which: The attribute indicating the function handle to obtain from underlying `Xform` objects.
            Typically either 'forward' or 'adjoint'.
        :return: An Image object as a result of applying forward or adjoint transformation to `im`.
        """
        # Ensure that we will be able to apply all transformers to the image
        assert (
            self.n_indices >= im.n_images
        ), f"Can process Image object of max depth {self.n_indices}. Got {im.n_images}."

        im_data = np.empty_like(im.asnumpy())

        # For each individual transformation
        for i, xform in enumerate(self.unique_xforms):
            # Get the indices corresponding to that transformation
            idx = np.flatnonzero(self.indices == i)
            # For the incoming Image object, find out which transformation indices are applicable
            idx = np.intersect1d(idx, indices)
            # For the transformation indices we found, find the indices in the Image object that we'll use
            im_data_indices = np.flatnonzero(np.isin(indices, idx))
            # Apply the transformation to the selected indices in the Image object
            if len(im_data_indices) > 0:
                fn_handle = getattr(xform, which)
                im_data[im_data_indices] = fn_handle(im[im_data_indices]).asnumpy()

        return Image(im_data)

    def _forward(self, im, indices):
        return self._indexed_operation(im, indices, "forward")


class LinearIndexedXform(IndexedXform, LinearXform):
    def _adjoint(self, im, indices):
        return self._indexed_operation(im, indices, "adjoint")


def _apply_xform(xform, im, indices, adjoint=False):
    """
    A simple global function (i.e. not a method) that is capable of being cached by joblib.Memory object's `cache`
    method.
    """
    if not adjoint:
        logger.debug("  Applying " + str(xform))
        return xform.forward(im, indices=indices)
    else:
        logger.debug("  Applying Adjoint " + str(xform))
        return xform.adjoint(im, indices=indices)


class Pipeline(Xform):
    """
    A `Pipeline` is a `Xform` made up of individual transformation steps (i.e. multiple `Xform` objects).
    The `Pipeline`, just like any other `Xform`, can be run in the `forward` (or `adjoint` mode, in the case of a
    `LinearPipeline`.

    In addition to keeping client-side code clean, a major advantage of `Pipeline` is that individual steps of the
    pipeline can be cached transparently by the `Pipeline`, providing significant performance advantages for steps that
    are performed repeatedly (especially during development while setting up these pipelines) on any Image/Xform pair.
    This caching uses `joblib.Memory` object behind the scenes, but is disabled by default.
    """

    def __init__(self, xforms=None, memory=None):
        """
        Initialize a `Pipeline` with `Xform` objects.

        :param xforms: An iterable of Xform objects to use in the Pipeline.
        :param memory: None for no caching (default), or the location of a directory to use to cache steps of the
            pipeline.
        """
        self.xforms = xforms or []
        self.memory = memory
        self.active = True

    def __str__(self):
        return "Apply pipeline: " + " ".join([f"{xform}" for xform in self.xforms])

    def add_xform(self, xform):
        """
        Add a single `Xform` object at the end of the pipeline.

        :param xform: A `Xform` object.
        :return: None
        """
        self.xforms.append(xform)

    def add_xforms(self, xforms):
        """
        Add multiple `Xform` objects at the end of the pipeline.

        :param xform: An iterable of `Xform` objects.
        :return: None
        """
        self.xforms.extend(xforms)

    def reset(self):
        """
        Reset the pipeline by removing all `Xform`s.

        :return: None
        """
        self.xforms = []

    def _forward(self, im, indices):
        memory = Memory(location=self.memory, verbose=0)
        _apply_transform_cached = memory.cache(_apply_xform)

        logger.debug("Applying forward transformations in pipeline")
        for xform in self.xforms:
            im = _apply_transform_cached(xform, im, indices, False)
        logger.debug("All forward transformations applied")

        return im


class LinearPipeline(Pipeline, LinearXform):
    def _adjoint(self, im, indices):
        memory = Memory(location=self.memory, verbose=0)
        _apply_transform_cached = memory.cache(_apply_xform)

        logger.debug("Applying adjoint transformations in pipeline")
        for xform in self.xforms[::-1]:
            im = _apply_transform_cached(xform, im, indices, True)
        logger.debug("All adjoint transformations applied")

        return im



import logging

import numpy as np
from numpy import pi

from aspire.basis import FBBasis3D
from aspire.basis.basis_utils import lgwt, norm_assoc_legendre, sph_bessel
from aspire.nufft import anufft, nufft
from aspire.utils.matlab_compat import m_flatten, m_reshape

logger = logging.getLogger(__name__)


class FFBBasis3D(FBBasis3D):
    """
    Define a derived class for fast spherical Harmonics Bessel basis expanding 3D volumes

    # TODO: Methods that return dictionaries should return useful objects instead

    """

    def _build(self):
        """
        Build the internal data structure for 3D Fourier-Bessel basis
        """
        logger.info(
            "Expanding 3D map in a frequency-domain Fourier–Bessel"
            " basis using the fast method."
        )

        # set cutoff values
        self.rcut = self.nres / 2
        self.kcut = 0.5

        # get upper bound of zeros, ells, and ks  of Bessel functions
        self._calc_k_max()

        # calculate total number of basis functions
        self.count = sum(self.k_max * (2 * np.arange(0, self.ell_max + 1) + 1))

        # generate 1D indices for basis functions
        self._indices = self.indices()

        # precompute the basis functions in 3D grids
        self._precomp = self._precomp()

        # get normalized factors
        self._norms = self.norms()

    def _precomp(self):
        """
        Precomute the basis functions on a polar Fourier 3D grid

        Gaussian quadrature points and weights are also generated
        in radical and phi dimensions.
        """
        n_r = int(np.ceil(4 * self.rcut * self.kcut))
        n_theta = int(2 * self.nres)
        n_phi = int(2 * self.nres + 1)

        r, wt_r = lgwt(n_r, 0.0, self.kcut, dtype=self.dtype)
        z, wt_z = lgwt(n_phi, -1, 1, dtype=self.dtype)
        r = m_reshape(r, (n_r, 1))
        wt_r = m_reshape(wt_r, (n_r, 1))
        z = m_reshape(z, (n_phi, 1))
        wt_z = m_reshape(wt_z, (n_phi, 1))
        phi = np.arccos(z)
        wt_phi = wt_z
        theta = 2 * pi * np.arange(n_theta, dtype=self.dtype).T / (2 * n_theta)
        theta = m_reshape(theta, (n_theta, 1))

        # evaluate basis function in the radial dimension
        radial_wtd = np.zeros(
            shape=(n_r, np.max(self.k_max), self.ell_max + 1), dtype=self.dtype
        )
        for ell in range(0, self.ell_max + 1):
            k_max_ell = self.k_max[ell]
            rmat = r * self.r0[ell][0:k_max_ell].T / self.kcut
            radial_ell = np.zeros_like(rmat)
            for ik in range(0, k_max_ell):
                radial_ell[:, ik] = sph_bessel(ell, rmat[:, ik])
            nrm = np.abs(sph_bessel(ell + 1, self.r0[ell][0:k_max_ell].T) / 4)
            radial_ell = radial_ell / nrm
            radial_ell_wtd = r**2 * wt_r * radial_ell
            radial_wtd[:, 0:k_max_ell, ell] = radial_ell_wtd

        # evaluate basis function in the phi dimension
        ang_phi_wtd_even = []
        ang_phi_wtd_odd = []
        for m in range(0, self.ell_max + 1):
            n_even_ell = int(
                np.floor((self.ell_max - m) / 2)
                + 1
                - np.mod(self.ell_max, 2) * np.mod(m, 2)
            )
            n_odd_ell = int(self.ell_max - m + 1 - n_even_ell)
            phi_wtd_m_even = np.zeros((n_phi, n_even_ell), dtype=phi.dtype)
            phi_wtd_m_odd = np.zeros((n_phi, n_odd_ell), dtype=phi.dtype)

            ind_even = 0
            ind_odd = 0
            for ell in range(m, self.ell_max + 1):
                phi_m_ell = norm_assoc_legendre(ell, m, z)
                nrm_inv = np.sqrt(0.5 / pi)
                phi_m_ell = nrm_inv * phi_m_ell
                phi_wtd_m_ell = wt_phi * phi_m_ell
                if np.mod(ell, 2) == 0:
                    phi_wtd_m_even[:, ind_even] = phi_wtd_m_ell[:, 0]
                    ind_even = ind_even + 1
                else:
                    phi_wtd_m_odd[:, ind_odd] = phi_wtd_m_ell[:, 0]
                    ind_odd = ind_odd + 1

            ang_phi_wtd_even.append(phi_wtd_m_even)
            ang_phi_wtd_odd.append(phi_wtd_m_odd)

        # evaluate basis function in the theta dimension
        ang_theta = np.zeros((n_theta, 2 * self.ell_max + 1), dtype=theta.dtype)

        ang_theta[:, 0 : self.ell_max] = np.sqrt(2) * np.sin(
            theta @ m_reshape(np.arange(self.ell_max, 0, -1), (1, self.ell_max))
        )
        ang_theta[:, self.ell_max] = np.ones(n_theta, dtype=theta.dtype)
        ang_theta[:, self.ell_max + 1 : 2 * self.ell_max + 1] = np.sqrt(2) * np.cos(
            theta @ m_reshape(np.arange(1, self.ell_max + 1), (1, self.ell_max))
        )

        ang_theta_wtd = (2 * pi / n_theta) * ang_theta

        theta_grid, phi_grid, r_grid = np.meshgrid(
            theta, phi, r, sparse=False, indexing="ij"
        )
        fourier_x = m_flatten(r_grid * np.cos(theta_grid) * np.sin(phi_grid))
        fourier_y = m_flatten(r_grid * np.sin(theta_grid) * np.sin(phi_grid))
        fourier_z = m_flatten(r_grid * np.cos(phi_grid))
        fourier_pts = (
            2
            * pi
            * np.vstack(
                (
                    fourier_z[np.newaxis, ...],
                    fourier_y[np.newaxis, ...],
                    fourier_x[np.newaxis, ...],
                )
            )
        )

        return {
            "radial_wtd": radial_wtd,
            "ang_phi_wtd_even": ang_phi_wtd_even,
            "ang_phi_wtd_odd": ang_phi_wtd_odd,
            "ang_theta_wtd": ang_theta_wtd,
            "fourier_pts": fourier_pts,
        }

    def _evaluate(self, v):
        """
        Evaluate coefficients in standard 3D coordinate basis from those in 3D FB basis

        :param v: A coefficient vector (or an array of coefficient vectors) in FB basis
            to be evaluated. The last dimension must equal `self.count`.
        :return x: The evaluation of the coefficient vector(s) `x` in standard 3D
            coordinate basis. This is an array whose last three dimensions equal
            `self.sz` and the remaining dimensions correspond to `v`.
        """
        # roll dimensions of v
        sz_roll = v.shape[:-1]
        v = v.reshape((-1, self.count))

        # get information on polar grids from precomputed data
        n_theta = np.size(self._precomp["ang_theta_wtd"], 0)
        n_phi = np.size(self._precomp["ang_phi_wtd_even"][0], 0)
        n_r = np.size(self._precomp["radial_wtd"], 0)

        # number of 3D image samples
        n_data = v.shape[0]

        u_even = np.zeros(
            (
                n_r,
                int(2 * self.ell_max + 1),
                n_data,
                int(np.floor(self.ell_max / 2) + 1),
            ),
            dtype=v.dtype,
        )
        u_odd = np.zeros(
            (n_r, int(2 * self.ell_max + 1), n_data, int(np.ceil(self.ell_max / 2))),
            dtype=v.dtype,
        )

        # go through each basis function and find corresponding coefficient
        # evaluate the radial parts
        for ell in range(0, self.ell_max + 1):
            k_max_ell = self.k_max[ell]
            radial_wtd = self._precomp["radial_wtd"][:, 0:k_max_ell, ell]

            ind = self._indices["ells"] == ell

            v_ell = m_reshape(v[:, ind].T, (k_max_ell, (2 * ell + 1) * n_data))
            v_ell = radial_wtd @ v_ell
            v_ell = m_reshape(v_ell, (n_r, 2 * ell + 1, n_data))

            if np.mod(ell, 2) == 0:
                u_even[
                    :,
                    int(self.ell_max - ell) : int(self.ell_max + ell + 1),
                    :,
                    int(ell / 2),
                ] = v_ell
            else:
                u_odd[
                    :,
                    int(self.ell_max - ell) : int(self.ell_max + ell + 1),
                    :,
                    int((ell - 1) / 2),
                ] = v_ell

        u_even = np.transpose(u_even, (3, 0, 1, 2))
        u_odd = np.transpose(u_odd, (3, 0, 1, 2))
        w_even = np.zeros((n_phi, n_r, n_data, 2 * self.ell_max + 1), dtype=v.dtype)
        w_odd = np.zeros((n_phi, n_r, n_data, 2 * self.ell_max + 1), dtype=v.dtype)

        # evaluate the phi parts
        for m in range(0, self.ell_max + 1):
            ang_phi_wtd_m_even = self._precomp["ang_phi_wtd_even"][m]
            ang_phi_wtd_m_odd = self._precomp["ang_phi_wtd_odd"][m]

            n_even_ell = np.size(ang_phi_wtd_m_even, 1)
            n_odd_ell = np.size(ang_phi_wtd_m_odd, 1)

            if m == 0:
                sgns = (1,)
            else:
                sgns = (1, -1)

            for sgn in sgns:
                end = np.size(u_even, 0)
                u_m_even = u_even[end - n_even_ell : end, :, self.ell_max + sgn * m, :]
                end = np.size(u_odd, 0)
                u_m_odd = u_odd[end - n_odd_ell : end, :, self.ell_max + sgn * m, :]

                u_m_even = m_reshape(u_m_even, (n_even_ell, n_r * n_data))
                u_m_odd = m_reshape(u_m_odd, (n_odd_ell, n_r * n_data))

                w_m_even = ang_phi_wtd_m_even @ u_m_even
                w_m_odd = ang_phi_wtd_m_odd @ u_m_odd

                w_m_even = m_reshape(w_m_even, (n_phi, n_r, n_data))
                w_m_odd = m_reshape(w_m_odd, (n_phi, n_r, n_data))

                w_even[:, :, :, self.ell_max + sgn * m] = w_m_even
                w_odd[:, :, :, self.ell_max + sgn * m] = w_m_odd

        w_even = np.transpose(w_even, (3, 0, 1, 2))
        w_odd = np.transpose(w_odd, (3, 0, 1, 2))
        u_even = w_even
        u_odd = w_odd

        u_even = m_reshape(u_even, (2 * self.ell_max + 1, n_phi * n_r * n_data))
        u_odd = m_reshape(u_odd, (2 * self.ell_max + 1, n_phi * n_r * n_data))

        # evaluate the theta parts
        w_even = self._precomp["ang_theta_wtd"] @ u_even
        w_odd = self._precomp["ang_theta_wtd"] @ u_odd

        pf = w_even + 1j * w_odd
        pf = m_reshape(pf, (n_theta * n_phi * n_r, n_data))
        pf = np.moveaxis(pf, 0, -1)

        # perform inverse non-uniformly FFT transformation back to 3D rectangular coordinates
        freqs = m_reshape(self._precomp["fourier_pts"], (3, n_r * n_theta * n_phi))
        x = anufft(pf, freqs, self.sz, real=True)

        # Roll, return the x with the last three dimensions as self.sz
        # Higher dimensions should be like v.
        x = x.reshape((*sz_roll, *self.sz))
        return x

    def _evaluate_t(self, x):
        """
        Evaluate coefficient in FB basis from those in standard 3D coordinate basis

        :param x: The coefficient array in the standard 3D coordinate basis
            to be evaluated. The last three dimensions must equal `self.sz`.
        :return: The evaluation of the coefficient array `x` in the FB basis.
            This is an array of vectors whose last dimension equals
            `self.count` and whose remaining dimensions correspond to higher
            dimensions of `x`.
        """
        # roll dimensions
        sz_roll = x.shape[:-3]
        x = x.reshape((-1, *self.sz))

        n_data = x.shape[0]
        n_r = np.size(self._precomp["radial_wtd"], 0)
        n_phi = np.size(self._precomp["ang_phi_wtd_even"][0], 0)
        n_theta = np.size(self._precomp["ang_theta_wtd"], 0)

        # resamping x in a polar Fourier gird using nonuniform discrete Fourier transform
        pf = nufft(x, self._precomp["fourier_pts"])

        pf = m_reshape(pf.T, (n_theta, n_phi * n_r * n_data))

        # evaluate the theta parts
        u_even = self._precomp["ang_theta_wtd"].T @ np.real(pf)
        u_odd = self._precomp["ang_theta_wtd"].T @ np.imag(pf)

        u_even = m_reshape(u_even, (2 * self.ell_max + 1, n_phi, n_r, n_data))
        u_odd = m_reshape(u_odd, (2 * self.ell_max + 1, n_phi, n_r, n_data))

        u_even = np.transpose(u_even, (1, 2, 3, 0))
        u_odd = np.transpose(u_odd, (1, 2, 3, 0))

        w_even = np.zeros(
            (int(np.floor(self.ell_max / 2) + 1), n_r, 2 * self.ell_max + 1, n_data),
            dtype=x.dtype,
        )
        w_odd = np.zeros(
            (int(np.ceil(self.ell_max / 2)), n_r, 2 * self.ell_max + 1, n_data),
            dtype=x.dtype,
        )

        # evaluate the phi parts
        for m in range(0, self.ell_max + 1):
            ang_phi_wtd_m_even = self._precomp["ang_phi_wtd_even"][m]
            ang_phi_wtd_m_odd = self._precomp["ang_phi_wtd_odd"][m]

            n_even_ell = np.size(ang_phi_wtd_m_even, 1)
            n_odd_ell = np.size(ang_phi_wtd_m_odd, 1)

            if m == 0:
                sgns = (1,)
            else:
                sgns = (1, -1)

            for sgn in sgns:
                u_m_even = u_even[:, :, :, self.ell_max + sgn * m]
                u_m_odd = u_odd[:, :, :, self.ell_max + sgn * m]

                u_m_even = m_reshape(u_m_even, (n_phi, n_r * n_data))
                u_m_odd = m_reshape(u_m_odd, (n_phi, n_r * n_data))

                w_m_even = ang_phi_wtd_m_even.T @ u_m_even
                w_m_odd = ang_phi_wtd_m_odd.T @ u_m_odd

                w_m_even = m_reshape(w_m_even, (n_even_ell, n_r, n_data))
                w_m_odd = m_reshape(w_m_odd, (n_odd_ell, n_r, n_data))
                end = np.size(w_even, 0)
                w_even[end - n_even_ell : end, :, self.ell_max + sgn * m, :] = w_m_even
                end = np.size(w_odd, 0)
                w_odd[end - n_odd_ell : end, :, self.ell_max + sgn * m, :] = w_m_odd

        w_even = np.transpose(w_even, (1, 2, 3, 0))
        w_odd = np.transpose(w_odd, (1, 2, 3, 0))

        # evaluate the radial parts
        v = np.zeros((n_data, self.count), dtype=x.dtype)
        for ell in range(0, self.ell_max + 1):
            k_max_ell = self.k_max[ell]
            radial_wtd = self._precomp["radial_wtd"][:, 0:k_max_ell, ell]

            if np.mod(ell, 2) == 0:
                v_ell = w_even[
                    :,
                    int(self.ell_max - ell) : int(self.ell_max + 1 + ell),
                    :,
                    int(ell / 2),
                ]
            else:
                v_ell = w_odd[
                    :,
                    int(self.ell_max - ell) : int(self.ell_max + 1 + ell),
                    :,
                    int((ell - 1) / 2),
                ]

            v_ell = m_reshape(v_ell, (n_r, (2 * ell + 1) * n_data))

            v_ell = radial_wtd.T @ v_ell

            v_ell = m_reshape(v_ell, (k_max_ell * (2 * ell + 1), n_data))

            # TODO: Fix this to avoid lookup each time.
            ind = self._indices["ells"] == ell
            v[:, ind] = v_ell.T

        # Roll dimensions, last dimension should be self.count,
        # Higher dimensions like x.
        v = v.reshape((*sz_roll, self.count))
        return v



"""
Miscellaneous Utilities that relate to logging.
"""
import logging
import os.path
import subprocess
from collections import defaultdict

import tqdm as _tqdm

from aspire import config

logger = logging.getLogger(__name__)

LOGGING_LEVEL_NAMES = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


def get_full_version():
    """
    Get as much version information as we can, including git info (if applicable)
    This method should never raise exceptions!

    :return: A version number in the form:
        <maj>.<min>.<bld>
            If we're running as a package distributed through setuptools
        <maj>.<min>.<bld>.<rev>
            If we're running as a 'regular' python source folder, possibly locally modified

            <rev> is one of:
                'src': The package is running as a source folder
                <git_tag> or <git_rev> or <git_rev>-dirty: A git tag or commit revision, possibly followed by a suffix
                    '-dirty' if source is modified locally
                'x':   The revision cannot be determined

    """
    import aspire

    full_version = aspire.__version__
    rev = None
    try:
        path = aspire.__path__[0]
        if os.path.isdir(path):
            # We have a package folder where we can get git information
            try:
                rev = (
                    subprocess.check_output(
                        ["git", "describe", "--tags", "--always", "--dirty"],
                        stderr=subprocess.STDOUT,
                        cwd=path,
                    )
                    .decode("utf-8")
                    .strip()
                )
            except (FileNotFoundError, subprocess.CalledProcessError):
                # no git or not a git repo? assume 'src'
                rev = "src"
    except Exception:  # nopep8  # noqa: E722
        # Something unexpected happened - rev number defaults to 'x'
        rev = "x"

    if rev is not None:
        full_version += f".{rev}"

    return full_version


def tqdm(*args, **kwargs):
    """
    Wraps `tqdm.tqdm`, applying ASPIRE configuration.

    Currently setting `aspire.config['logging']['tqdm_disable']`
    true/false will disable/enable tqdm progress bars.
    """

    disable = config["logging"]["tqdm_disable"] or (
        getConsoleLoggingLevel() not in ["DEBUG", "INFO"]
    )
    return _tqdm.tqdm(*args, **kwargs, disable=disable)


def trange(*args, **kwargs):
    """
    Wraps `tqdm.trange`, applying ASPIRE configuration.

    Currently setting `aspire.config['logging']['tqdm_disable']`
    true/false will disable/enable tqdm progress bars.
    """

    disable = config["logging"]["tqdm_disable"] or (
        getConsoleLoggingLevel() not in ["DEBUG", "INFO"]
    )
    return _tqdm.trange(*args, **kwargs, disable=disable)


def setConsoleLoggingLevel(level_name):
    """
    Dynamically sets the console logging level by setting the level of the root logger's StreamHandler to `level_name`.
    Note this will supersede the `logging.console_level` option stored in ASPIRE's configuration file.

    :param level_name: One of "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL".
    """
    if level_name not in LOGGING_LEVEL_NAMES:
        raise ValueError(
            f"{level_name} not a recognized logging level. Must be one of {LOGGING_LEVEL_NAMES}"
        )
    # handler list is ordered according to logging.conf
    stream_handler = logging.getLogger().handlers[0]
    stream_handler.setLevel(getattr(logging, level_name))


def getConsoleLoggingLevel():
    """
    Returns the Python logging level of the root logger's StreamHandler, i.e. console output.
    This is the same as the `logging.console_level` option in ASPIRE's configuration file unless
    the console logging level has been changed dynamically during a session. (e.g. by a CLI option)

    :return: The current console logging level name as a string. One of "DEBUG". "INFO", "WARNING",
    "ERROR", "CRITICAL".
    """
    # handler list is ordered according to logging.conf
    stream_handler = logging.getLogger().handlers[0]
    return logging.getLevelName(stream_handler.level)


def setFileLoggingLevel(level_name):
    """
    Dynamically sets the log file logging level by setting the level of the root logger's FileHandler to `level_name`.
    Note this will supersede the `logging.log_file_level` option stored in ASPIRE's configuration file.

    :param level_name: One of "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL".
    """
    if level_name not in LOGGING_LEVEL_NAMES:
        raise ValueError(
            f"{level_name} not a recognized logging level. Must be one of {LOGGING_LEVEL_NAMES}"
        )
    # handler list is ordered according to logging.conf
    file_handler = logging.getLogger().handlers[1]
    file_handler.setLevel(getattr(logging, level_name))


def getFileLoggingLevel():
    """
    Returns the Python logging level of the root logger's FileHandler, i.e. log file output.
    This is the same as the `logging.log_file_level` option in ASPIRE's configuration file
    unless the file logging level has been changed dynamically during a session. (e.g. by a CLI option)

    :return: The current file logging level name as a string. One of "DEBUG", "INFO", "WARNING",
    "ERROR", "CRITICAL".
    """
    # handler list is ordered according to logging.conf
    file_handler = logging.getLogger().handlers[1]
    return logging.getLevelName(file_handler.level)


class LogFilterByCount:
    """
    Provide a context manager for filtering repetitive log messages.
    """

    # msg_cache is intentionally shared by all instances of class.
    # msg_cache is map hash(str(msg)) ~~> count.
    msg_cache = defaultdict(int)

    def __init__(self, logger: logging.Logger, max_count: int):
        """
        Initialize context manager based on `logger` and `max_count`.

        :param logger: `Logger` instance.
        :param max_count: Global limit for count of each message
            encountered inside context.
        """

        self._logger = logger
        self._max_count = max_count

    def filter(self, record):
        """
        Increment msg_cache for `record`.

        `filter` returns True when `seen` <= `count` for this context.
        True implies the logger will pass the message.

        :param record: Log record.  Will be reduced by hash(str()),
        :return: Boolean
        """

        # Log messages can be arbitrarily long,
        #   convert to a hash(str())
        msg = hash(str(record.msg))

        # Increment seen count in cache.
        self.msg_cache[msg] += 1
        seen = self.msg_cache[msg]

        return seen <= self._max_count

    def __enter__(self):
        self._logger.addFilter(self)

    def __exit__(self, *args):
        self._logger.removeFilter(self)



import logging
from warnings import catch_warnings, filterwarnings, warn

import matplotlib.pyplot as plt
import mrcfile
import numpy as np
from scipy.linalg import lstsq

import aspire.volume
from aspire.nufft import anufft
from aspire.numeric import fft, xp
from aspire.utils import crop_pad_2d, grid_2d
from aspire.utils.matrix import anorm

logger = logging.getLogger(__name__)


def normalize_bg(imgs, bg_radius=1.0, do_ramp=True):
    """
    Normalize backgrounds and apply to a stack of images

    :param imgs: A stack of images in N-by-L-by-L array
    :param bg_radius: Radius cutoff to be considered as background (in image size)
    :param do_ramp: When it is `True`, fit a ramping background to the data
            and subtract. Namely perform normalization based on values from each image.
            Otherwise, a constant background level from all images is used.
    :return: The modified images
    """
    if imgs.ndim > 3:
        raise NotImplementedError(
            "`normalize_bg` is currently limited to 1D image stacks."
        )

    L = imgs.shape[-1]
    input_dtype = imgs.dtype
    grid = grid_2d(L, indexing="yx", dtype=input_dtype)
    mask = grid["r"] > bg_radius

    if do_ramp:
        # Create matrices and reshape the background mask
        # for fitting a ramping background
        ramp_mask = np.vstack(
            (
                grid["x"][mask].flatten(),
                grid["y"][mask].flatten(),
                np.ones(grid["y"][mask].flatten().size, dtype=input_dtype),
            )
        ).T
        ramp_all = np.vstack(
            (
                grid["x"].flatten(),
                grid["y"].flatten(),
                np.ones(L * L, dtype=input_dtype),
            )
        ).T
        mask_reshape = mask.reshape((L * L))
        imgs = imgs.reshape((-1, L * L))

        # Fit a ramping background and apply to images
        coeff = lstsq(ramp_mask, imgs[:, mask_reshape].T)[0]  # RCOPT
        imgs = imgs - (ramp_all @ coeff).T  # RCOPT
        imgs = imgs.reshape((-1, L, L))

    # Apply mask images and calculate mean and std values of background
    imgs_masked = imgs * mask
    denominator = np.sum(mask)
    first_moment = np.sum(imgs_masked, axis=(1, 2)) / denominator
    second_moment = np.sum(imgs_masked**2, axis=(1, 2)) / denominator
    mean = first_moment.reshape(-1, 1, 1)
    variance = second_moment.reshape(-1, 1, 1) - mean**2
    std = np.sqrt(variance)

    return (imgs - mean) / std


class Image:
    def __init__(self, data, dtype=None):
        """
        A stack of one or more images.

        This is a wrapper of numpy.ndarray which provides methods
        for common processing tasks.

        The stack can be multidimensional with `n_images` equal
        to the product of the stack dimensions.  Singletons will be
        expanded into a stack with one entry.

        The last two axes represent the image size,
        and are checked to be square.

        :param data: Numpy array containing image data with shape
            `(..., resolution, resolution)`.
        :param dtype: Optionally cast `data` to this dtype.
            Defaults to `data.dtype`.

        :return: Image instance holding `data`.
        """

        if not isinstance(data, np.ndarray):
            raise ValueError("Image should be instantiated with an ndarray")

        if data.ndim < 2:
            raise ValueError(
                "Image data should be ndarray with shape (N1...)xLxL or LxL."
            )
        elif data.ndim == 2:
            data = np.expand_dims(data, axis=0)

        if dtype is None:
            self.dtype = data.dtype
        else:
            self.dtype = np.dtype(dtype)

        if not data.shape[-1] == data.shape[-2]:
            raise ValueError("Only square ndarrays are supported.")

        self._data = data.astype(self.dtype, copy=False)
        self.ndim = self._data.ndim
        self.shape = self._data.shape
        self.stack_ndim = self._data.ndim - 2
        self.stack_shape = self._data.shape[:-2]
        self.n_images = np.prod(self.stack_shape)
        self.resolution = self._data.shape[-1]

        # Numpy interop
        # https://numpy.org/devdocs/user/basics.interoperability.html#the-array-interface-protocol
        self.__array_interface__ = self._data.__array_interface__
        self.__array__ = self._data

    @property
    def res(self):
        warn(
            "`Image.res` will be deprecated in favor or Image.resolution in an upcoming release.",
            DeprecationWarning,
        )
        return self.resolution

    def _check_key_dims(self, key):
        if isinstance(key, tuple) and (len(key) > self._data.ndim):
            raise ValueError(
                f"Image stack_dim is {self.stack_ndim}, slice length must be =< {self.ndim}"
            )

    def __getitem__(self, key):
        self._check_key_dims(key)
        return self.__class__(self._data[key])

    def __setitem__(self, key, value):
        self._check_key_dims(key)
        self._data[key] = value

    def stack_reshape(self, *args):
        """
        Reshape the stack axis.

        :*args: Integer(s) or tuple describing the intended shape.

        :returns: Image instance
        """

        # If we're passed a tuple, use that
        if len(args) == 1 and isinstance(args[0], tuple):
            shape = args[0]
        else:
            # Otherwise use the variadic args
            shape = args

        # Sanity check the size
        if shape != (-1,) and np.prod(shape) != self.n_images:
            raise ValueError(
                f"Number of images {self.n_images} cannot be reshaped to {shape}."
            )

        return self.__class__(self._data.reshape(*shape, *self._data.shape[-2:]))

    def __add__(self, other):
        if isinstance(other, Image):
            other = other._data

        return self.__class__(self._data + other)

    def __sub__(self, other):
        if isinstance(other, Image):
            other = other._data

        return self.__class__(self._data - other)

    def __mul__(self, other):
        if isinstance(other, Image):
            other = other._data

        return self.__class__(self._data * other)

    def __neg__(self):
        return self.__class__(-self._data)

    def sqrt(self):
        return self.__class__(np.sqrt(self._data))

    @property
    def T(self):
        """
        Abbreviation for transpose.

        :return: Image instance.
        """
        return self.transpose()

    def transpose(self):
        """
        Returns a new Image instance with image data axes transposed.

        :return: Image instance.
        """
        original_stack_shape = self.stack_shape

        im = self.stack_reshape(-1)
        imt = np.transpose(im._data, (0, -1, -2))

        return self.__class__(imt).stack_reshape(original_stack_shape)

    def flip(self, axis=-2):
        """
        Flip image stack data along axis using numpy.flip().

        :param axis: Optionally specify axis as integer or tuple.
            Defaults to axis=-2.

        :return: Image instance.
        """
        # Convert integer to tuple, so we can always loop.
        if isinstance(axis, int):
            axis = (axis,)

        # Check we are not attempting to flip any stack axis.
        for ax in axis:
            ax = ax % self.ndim  # modulo [0, ndim)
            if ax < self.stack_ndim:
                raise ValueError(
                    f"Cannot flip axis {ax}: stack axis. Did you mean {ax-3}?"
                )

        return self.__class__(np.flip(self._data, axis))

    def __repr__(self):
        msg = f"{self.n_images} {self.dtype} images arranged as a {self.stack_shape} stack"
        msg += f" each of size {self.resolution}x{self.resolution}."
        return msg

    def asnumpy(self):
        """
        Return image data as a (<stack>, resolution, resolution)
        read-only array view.

        :return: read-only ndarray view
        """

        view = self._data.view()
        view.flags.writeable = False
        return view

    def copy(self):
        return self.__class__(self._data.copy())

    def shift(self, shifts):
        """
        Translate image by shifts. This method returns a new Image.

        :param shifts: An array of size n-by-2 specifying the shifts in pixels.
            Alternatively, it can be a column vector of length 2, in which case
            the same shifts is applied to each image.
        :return: The Image translated by the shifts, with periodic boundaries.
        """
        if shifts.ndim == 1:
            shifts = shifts[np.newaxis, :]

        n_shifts = shifts.shape[0]

        if not shifts.shape[1] == 2:
            raise ValueError("Input shifts must be of shape (n_images, 2) or (1, 2).")
        if not n_shifts == 1 and not n_shifts == self.n_images:
            raise ValueError(
                "The number of shifts must be 1 or equal to self.n_images."
            )

        return self._im_translate(shifts)

    def downsample(self, ds_res):
        """
        Downsample Image to a specific resolution. This method returns a new Image.

        :param ds_res: int - new resolution, should be <= the current resolution
            of this Image
        :return: The downsampled Image object.
        """

        original_stack_shape = self.stack_shape
        im = self.stack_reshape(-1)

        # compute FT with centered 0-frequency
        fx = fft.centered_fft2(im._data)
        # crop 2D Fourier transform for each image
        crop_fx = np.array([crop_pad_2d(fx[i], ds_res) for i in range(self.n_images)])
        # take back to real space, discard complex part, and scale
        out = np.real(fft.centered_ifft2(crop_fx)) * (
            ds_res**2 / self.resolution**2
        )

        return self.__class__(out).stack_reshape(original_stack_shape)

    def filter(self, filter):
        """
        Apply a `Filter` object to the Image and returns a new Image.

        :param filter: An object of type `Filter`.
        :return: A new filtered `Image` object.
        """
        original_stack_shape = self.stack_shape

        im = self.stack_reshape(-1)

        filter_values = filter.evaluate_grid(self.resolution)

        im_f = xp.asnumpy(fft.centered_fft2(xp.asarray(im._data)))

        # TODO: why are these different? Doesn't the broadcast work?
        if im_f.ndim > filter_values.ndim:
            im_f *= filter_values
        else:
            im_f = filter_values * im_f
        im = xp.asnumpy(fft.centered_ifft2(xp.asarray(im_f)))
        im = np.real(im)

        return self.__class__(im).stack_reshape(original_stack_shape)

    def rotate(self):
        raise NotImplementedError

    def save(self, mrcs_filepath, overwrite=False):
        if self.stack_ndim > 1:
            raise NotImplementedError("`save` is currently limited to 1D image stacks.")

        with mrcfile.new(mrcs_filepath, overwrite=overwrite) as mrc:
            # original input format (the image index first)
            mrc.set_data(self._data.astype(np.float32))

    def _im_translate(self, shifts):
        """
        Translate image by shifts
        :param im: An array of size n-by-L-by-L containing images to be translated.
        :param shifts: An array of size n-by-2 specifying the shifts in pixels.
            Alternatively, it can be a row vector of length 2, in which case the same shifts is applied to each image.
        :return: The images translated by the shifts, with periodic boundaries.
        """

        # Note original stack shape and flatten stack
        stack_shape = self.stack_shape
        im = self.stack_reshape(-1)._data

        if shifts.ndim == 1:
            shifts = shifts[np.newaxis, :]
        n_shifts = shifts.shape[0]

        assert shifts.shape[-1] == 2, "shifts must be nx2"

        assert (
            n_shifts == 1 or n_shifts == self.n_images
        ), "number of shifts must be 1 or match the number of images"
        # Cast shifts to this instance's internal dtype
        shifts = shifts.astype(self.dtype)

        L = self.resolution
        im_f = xp.asnumpy(fft.fft2(xp.asarray(im)))
        grid_shifted = fft.ifftshift(
            xp.asarray(np.ceil(np.arange(-L / 2, L / 2, dtype=self.dtype)))
        )
        grid_1d = xp.asnumpy(grid_shifted) * 2 * np.pi / L
        om_x, om_y = np.meshgrid(grid_1d, grid_1d, indexing="ij")

        phase_shifts_x = -shifts[:, 0].reshape((n_shifts, 1, 1))
        phase_shifts_y = -shifts[:, 1].reshape((n_shifts, 1, 1))

        phase_shifts = (
            om_x[np.newaxis, :, :] * phase_shifts_x
            + om_y[np.newaxis, :, :] * phase_shifts_y
        )
        mult_f = np.exp(-1j * phase_shifts)
        im_translated_f = im_f * mult_f
        im_translated = xp.asnumpy(fft.ifft2(xp.asarray(im_translated_f)))
        im_translated = np.real(im_translated)

        # Reshape to stack shape
        return self.__class__(im_translated).stack_reshape(stack_shape)

    def norm(self):
        return anorm(self._data)

    @property
    def size(self):
        # probably not needed, transition
        return np.size(self._data)

    def backproject(self, rot_matrices):
        """
        Backproject images along rotation
        :param im: An Image (stack) to backproject.
        :param rot_matrices: An n-by-3-by-3 array of rotation matrices \
        corresponding to viewing directions.

        :return: Volume instance corresonding to the backprojected images.
        """

        if self.stack_ndim > 1:
            raise NotImplementedError(
                "`Backprojection` is currently limited to 1D image stacks."
            )

        L = self.resolution

        assert (
            self.n_images == rot_matrices.shape[0]
        ), "Number of rotation matrices must match the number of images"

        # TODO: rotated_grids might as well give us correctly shaped array in the first place
        pts_rot = aspire.volume.rotated_grids(L, rot_matrices).astype(
            self.dtype, copy=False
        )
        pts_rot = pts_rot.reshape((3, -1))

        im_f = xp.asnumpy(fft.centered_fft2(xp.asarray(self._data))) / (L**2)
        if L % 2 == 0:
            im_f[:, 0, :] = 0
            im_f[:, :, 0] = 0

        im_f = im_f.flatten()

        vol = anufft(im_f, pts_rot[::-1], (L, L, L), real=True) / L

        return aspire.volume.Volume(vol)

    def show(self, columns=5, figsize=(20, 10), colorbar=True):
        """
        Plotting Utility Function.

        :param columns: Number of columns in a row of plots.
        :param figsize: Figure size in inches, consult `matplotlib.figure`.
        :param colorbar: Optionally plot colorbar to show scale.
            Defaults to True. Accepts `bool` or `dictionary`,
            where the dictionary is passed to `matplotlib.pyplot.colorbar`.
        """

        if self.stack_ndim > 1:
            raise NotImplementedError("`show` is currently limited to 1D image stacks.")

        # We never need more columns than images.
        columns = min(columns, self.n_images)

        # Create an empty colorbar options dictionary as needed.
        colorbar_opts = colorbar if isinstance(colorbar, dict) else dict()

        # Create a context manager for altering warnings
        with catch_warnings():
            # Filter off specific warning.
            # sphinx-gallery overrides to `agg` backend, but doesn't handle warning.
            filterwarnings(
                "ignore",
                category=UserWarning,
                message="Matplotlib is currently using agg, which is a"
                " non-GUI backend, so cannot show the figure.",
            )

            plt.figure(figsize=figsize)
            for i, im in enumerate(self.asnumpy()):
                plt.subplot(self.n_images // columns + 1, columns, i + 1)
                plt.imshow(im, cmap="gray")
                if colorbar:
                    plt.colorbar(**colorbar_opts)

            plt.show()


class CartesianImage(Image):
    def expand(self, basis):
        return BasisImage(basis)


class PolarImage(Image):
    def expand(self, basis):
        return BasisImage(basis)


class BispecImage(Image):
    def expand(self, basis):
        return BasisImage(basis)


class BasisImage(Image):
    def __init__(self, basis):
        self.basis = basis

    def evaluate(self):
        return CartesianImage()


class FBBasisImage(BasisImage):
    pass



import logging

from aspire.source.coordinates import BoxesCoordinateSource, CentersCoordinateSource
from aspire.source.image import ArrayImageSource, ImageSource, IndexedSource
from aspire.source.relion import RelionSource
from aspire.source.simulation import Simulation

logger = logging.getLogger(__name__)



import os
from threading import Lock

import pyfftw
import pyfftw.interfaces.scipy_fftpack as scipy_fft

from aspire.numeric.base_fft import FFT
from aspire.utils.types import complex_type

mutex = Lock()

_cpu_count = os.cpu_count()


def _workers(workers):
    if workers in (None, 0):
        raise ValueError("Workers must be specified")

    if workers < 0:
        # Borrow the idea from scipy for negative values of workers.
        # When workers is -1, we use all available threads.
        # Otherwise, (-workers-1) will be saved for other tasks.
        if workers >= -_cpu_count:
            workers += 1 + _cpu_count
        else:
            raise ValueError(
                f"Workers value out of range; got {workers}, "
                f"must not be less than {-_cpu_count}"
            )

    return workers


class PyfftwFFT(FFT):
    """
    Define a unified wrapper class for PyFFT functions

    To be consistent with Scipy FFT, not all arguments are included.
    """

    def fft(self, a, axis=-1, workers=-1):
        mutex.acquire()

        comp_type = complex_type(a.dtype)
        try:
            a_ = pyfftw.empty_aligned(a.shape, dtype=comp_type)
            b = pyfftw.empty_aligned(a.shape, dtype=comp_type)
            cls = pyfftw.FFTW(
                a_, b, axes=(axis,), direction="FFTW_FORWARD", threads=_workers(workers)
            )
            cls(a, b)
        finally:
            mutex.release()

        return b

    def ifft(self, a, axis=-1, workers=-1):
        mutex.acquire()

        comp_type = a.dtype
        try:
            a_ = pyfftw.empty_aligned(a.shape, dtype=comp_type)
            b = pyfftw.empty_aligned(a.shape, dtype=comp_type)
            cls = pyfftw.FFTW(
                a_,
                b,
                axes=(axis,),
                direction="FFTW_BACKWARD",
                threads=_workers(workers),
            )
            cls(a, b)
        finally:
            mutex.release()

        return b

    def fft2(self, a, axes=(-2, -1), workers=-1):
        # This is called by ApplePicker unit test using ThreadPoolExecutor.
        #   I don't believe this pyfftw call is actually threadsafe.
        #   Holding mutex here, I have not been able to reproduce the spurious
        #   segmentation fault on Linux.
        #   This still allows threading in the other areas of invoking code,
        #   presumably the parts which are IO bound.
        mutex.acquire()

        comp_type = complex_type(a.dtype)
        try:
            a_ = pyfftw.empty_aligned(a.shape, dtype=comp_type)
            b = pyfftw.empty_aligned(a.shape, dtype=comp_type)
            cls = pyfftw.FFTW(
                a_, b, axes=axes, direction="FFTW_FORWARD", threads=_workers(workers)
            )
            cls(a, b)
        finally:
            mutex.release()

        return b

    def ifft2(self, a, axes=(-2, -1), workers=-1):
        mutex.acquire()

        comp_type = a.dtype
        try:
            a_ = pyfftw.empty_aligned(a.shape, dtype=comp_type)
            b = pyfftw.empty_aligned(a.shape, dtype=comp_type)
            cls = pyfftw.FFTW(
                a_, b, axes=axes, direction="FFTW_BACKWARD", threads=_workers(workers)
            )
            cls(a, b)
        finally:
            mutex.release()

        return b

    def fftn(self, a, axes=None, workers=-1):
        axes = axes or tuple(range(a.ndim))
        comp_type = complex_type(a.dtype)

        mutex.acquire()
        try:
            a_ = pyfftw.empty_aligned(a.shape, dtype=comp_type)
            b = pyfftw.empty_aligned(a.shape, dtype=comp_type)
            cls = pyfftw.FFTW(
                a_, b, axes=axes, direction="FFTW_FORWARD", threads=_workers(workers)
            )
            cls(a, b)
        finally:
            mutex.release()

        return b

    def ifftn(self, a, axes=None, workers=-1):
        axes = axes or tuple(range(a.ndim))

        # FFTW_BACKWARD requires complex input array, cast as needed.
        # See https://pyfftw.readthedocs.io/en/latest/source/pyfftw/pyfftw.html#scheme-table
        comp_type = complex_type(a.dtype)
        a = a.astype(comp_type, copy=False)

        mutex.acquire()
        try:
            a_ = pyfftw.empty_aligned(a.shape, dtype=comp_type)
            b = pyfftw.empty_aligned(a.shape, dtype=comp_type)
            cls = pyfftw.FFTW(
                a_, b, axes=axes, direction="FFTW_BACKWARD", threads=_workers(workers)
            )
            cls(a, b)
        finally:
            mutex.release()

        return b

    def fftshift(self, a, axes=None):
        return scipy_fft.fftshift(a, axes=axes)

    def ifftshift(self, a, axes=None):
        return scipy_fft.ifftshift(a, axes=axes)



import logging

import numpy as np

from aspire.numeric import fft
from aspire.utils import roll_dim, unroll_dim, vec_to_vol, vecmat_to_volmat, vol_to_vec
from aspire.utils.matlab_compat import m_reshape
from aspire.volume import Volume

logger = logging.getLogger(__name__)


class Kernel:
    pass


class FourierKernel(Kernel):
    def __init__(self, kernel, centered):
        self.ndim = kernel.ndim
        self.kernel = kernel
        self.M = kernel.shape[0]
        self.dtype = kernel.dtype

        # TODO: `centered` should be populated based on how the object is constructed, not explicitly
        self._centered = centered

    def __add__(self, delta):
        """
        Add a tiny delta to the underlying kernel.

        :param delta: A scalar or an `ndarray` that can be broadcast to the `kernel` attribute of this object.
        :return: A new FourierKernel object with a modified kernel

        .. note::
            There is often a need to add a regularization parameter (a small positive value) to a FourierKernel object,
            to be able to use it within optimization loops. This operator allows one to use the FourierKernel object
            with the underlying 'kernel' attribute tweaked with a regularization parameter.
        """
        new_kernel = self.kernel + delta
        return FourierKernel(new_kernel, self._centered)

    def is_centered(self):
        return self._centered

    def circularize(self):
        logger.info("Circularizing kernel")
        kernel = np.real(fft.ifftn(self.kernel))
        kernel = fft.mdim_fftshift(kernel)

        for dim in range(self.ndim):
            logger.info(f"Circularizing dimension {dim}")
            kernel = self.circularize_1d(kernel, dim)

        xx = fft.fftn(fft.mdim_ifftshift(kernel))
        return xx

    def circularize_1d(self, kernel, dim):
        ndim = kernel.ndim
        sz = kernel.shape
        N = int(sz[dim] / 2)

        top, bottom = np.split(kernel, 2, axis=dim)

        # Multiplier for weighted average
        mult_shape = [1] * ndim
        mult_shape[dim] = N
        mult_shape = tuple(mult_shape)

        mult = m_reshape((np.arange(N, dtype=self.dtype) / N), mult_shape)
        kernel_circ = mult * top

        mult = m_reshape((np.arange(N, 0, -1, dtype=self.dtype) / N), mult_shape)
        kernel_circ += mult * bottom

        return fft.fftshift(kernel_circ, dim)

    def convolve_volume(self, x):
        """
        Convolve volume with kernel

        :param x: An N-by-N-by-N-by-... array of volumes to be convolved.
        :return: The original volumes convolved by the kernel with the same dimensions as before.
        """
        # Note, we can do better here, but some changes in `wts` already...
        if isinstance(x, Volume):
            x = x.asnumpy()[0]

        N = x.shape[0]
        kernel_f = self.kernel[..., np.newaxis]
        N_ker = kernel_f.shape[0]

        x, sz_roll = unroll_dim(x, 4)
        assert x.shape[0] == x.shape[1] == x.shape[2] == N, "Volumes in x must be cubic"
        assert kernel_f.shape[3] == 1, "Convolution kernel must be cubic"
        assert len(set(kernel_f.shape[:3])) == 1, "Convolution kernel must be cubic"

        is_singleton = x.shape[3] == 1

        if is_singleton:
            pad_width = [(0, N_ker - N)] * 3
            x = np.pad(x[..., 0], pad_width)
            x = fft.fftn(x)[..., np.newaxis]
        else:
            raise NotImplementedError("not yet")

        x = x * kernel_f

        if is_singleton:
            x[..., 0] = np.real(fft.ifftn(x[..., 0]))
            x = x[:N, :N, :N, :]
        else:
            raise NotImplementedError("not yet")

        x = roll_dim(x, sz_roll)

        return np.real(x)

    def convolve_volume_matrix(self, x):
        """
        Convolve volume matrix with kernel

        :param x: An N-by-...-by-N (6 dimensions) volume matrix to be convolved.
        :return: The original volume matrix convolved by the kernel with the same dimensions as before.
        """
        shape = x.shape
        N = shape[0]
        kernel_f = self.kernel
        assert (
            len(set(shape[i] for i in range(5))) == 1
        ), "Volume matrix must be cubic and square"

        # TODO from MATLAB code: Deal with rolled dimensions
        N_ker = kernel_f.shape[0]

        # Note from MATLAB code:
        # Order is important here.  It's about 20% faster to run from 1 through 6 compared with 6 through 1.
        # TODO: Experiment with fft axis order
        for i in range(6):
            _pad_width = [(0, 0)] * 6
            _pad_width[i] = (0, N_ker - N)
            x = fft.fft(np.pad(x, _pad_width), axis=i)

        x *= kernel_f

        indices = list(range(N))
        for i in range(5, -1, -1):
            x = fft.ifft(x, axis=i)
            x = x.take(indices, axis=i)

        return np.real(x)

    def toeplitz(self, L=None):
        """
        Compute the 3D Toeplitz matrix corresponding to this Fourier Kernel

        :param L: The size of the volumes to be convolved (default M/2, where the dimensions of this Fourier Kernel
            are MxMxM
        :return: An six-dimensional Toeplitz matrix of size L describing the convolution of a volume with this kernel
        """
        if L is None:
            L = int(self.M / 2)

        A = np.eye(L**3, dtype=self.dtype)
        for i in range(L**3):
            A[:, i] = np.real(vol_to_vec(self.convolve_volume(vec_to_vol(A[:, i]))))

        A = vecmat_to_volmat(A)
        return A



from .averager2d import (
    AligningAverager2D,
    Averager2D,
    BFRAverager2D,
    BFSRAverager2D,
    BFSReddyChatterjiAverager2D,
    EMAverager2D,
    FTKAverager2D,
    ReddyChatterjiAverager2D,
)
from .class2d import Class2D
from .class_selection import (
    BandedSNRImageQualityFunction,
    BumpWeightedImageQualityMixin,
    BumpWeightedVarianceImageQualityFunction,
    ClassSelector,
    DistanceClassSelector,
    GlobalClassSelector,
    GlobalWithRepulsionClassSelector,
    NeighborVarianceClassSelector,
    NeighborVarianceWithRepulsionClassSelector,
    RampWeightedImageQualityMixin,
    RampWeightedVarianceImageQualityFunction,
    RandomClassSelector,
    TopClassSelector,
    VarianceImageQualityFunction,
    WeightedImageQualityMixin,
)
from .rir_class2d import RIRClass2D



import logging

import numpy as np
from scipy.sparse.linalg import LinearOperator, cg

from aspire.image import Image
from aspire.utils import mdim_mat_fun_conj
from aspire.volume import Volume

logger = logging.getLogger(__name__)


class Basis:
    """
    Define a base class for expanding 2D particle images and 3D structure volumes

    """

    def __init__(self, size, ell_max=None, dtype=np.float32):
        """
        Initialize an object for the base of basis class

        :param size: The size of the vectors for which to define the basis.
            Currently only square images and cubic volumes are supported.
        :param ell_max: The maximum order ell of the basis elements. If no input
            (= None), it will be set to np.Inf and the basis includes all
            ell such that the resulting basis vectors are concentrated
            below the Nyquist frequency (default Inf).
        """
        if ell_max is None:
            ell_max = np.inf

        ndim = len(size)
        nres = size[0]
        self.sz = size
        self.nres = nres
        self.count = 0
        self.ell_max = ell_max
        self.ndim = ndim
        if self.ndim == 2:
            self._cls = Image
        elif self.ndim == 3:
            self._cls = Volume
        else:
            raise RuntimeError("Basis ndim must be 2 or 3")
        self.dtype = np.dtype(dtype)
        if self.dtype not in (np.float32, np.float64):
            raise NotImplementedError(
                "Currently only implemented for float32 and float64 types"
            )
        # dtype of coefficients is the same as self.dtype for real bases
        # subclasses with complex coefficients override this attribute
        self.coefficient_dtype = self.dtype

        self._build()

    def _build(self):
        """
        Build the internal data structure to represent basis
        """
        raise NotImplementedError("subclasses must implement this")

    def indices(self):
        """
        Create the indices for each basis function
        """
        raise NotImplementedError("subclasses must implement this")

    def _precomp(self):
        """
        Precompute the basis functions at defined sample points
        """
        raise NotImplementedError("subclasses must implement this")

    def norms(self):
        """
        Calculate the normalized factors of basis functions
        """
        raise NotImplementedError("subclasses must implement this")

    def evaluate(self, v):
        """
        Evaluate coefficient vector in basis

        :param v: A coefficient vector (or an array of coefficient vectors)
            to be evaluated. The first dimension must correspond to the number of
            coefficient vectors, while the second must correspond to `self.count`
        :return: The evaluation of the coefficient vector(s) `v` for this basis.
            This is an Image or a Volume object containing one image/volume for each
            coefficient vector, and of size `self.sz`.
        """
        if v.dtype != self.coefficient_dtype:
            logger.warning(
                f"{self.__class__.__name__}::evaluate"
                f" Inconsistent dtypes v: {v.dtype} self coefficient dtype: {self.coefficient_dtype}"
            )

        # Flatten stack, ndim is wrt Basis (2 or 3)
        stack_shape = v.shape[:-1]
        v = v.reshape(-1, self.count)
        # Compute the transform
        x = self._evaluate(v)
        # Restore stack shape
        x = x.reshape(*stack_shape, *self.sz)

        # Return the appropriate class
        return self._cls(x)

    def _evaluate(self, v):
        raise NotImplementedError("subclasses must implement this")

    def evaluate_t(self, v):
        """
        Evaluate coefficient in dual basis

        :param v: An Image or Volume object whose size matches `self.sz`.
        :return: The evaluation of the Image or Volume object `v` in the dual
            basis of `basis`.
            This is an array of vectors whose first dimension equals the number of
            images/volumes in `v`. and whose second dimension is `self.count`.
        """
        if v.dtype != self.dtype:
            logger.warning(
                f"{self.__class__.__name__}::evaluate_t"
                f" Inconsistent dtypes v: {v.dtype} self: {self.dtype}"
            )

        if not isinstance(v, self._cls):
            logger.warning(
                f"{self.__class__.__name__}::evaluate_t"
                f" passed numpy array instead of {self._cls}."
            )
        else:
            v = v.asnumpy()

        # Flatten stack, ndim is wrt Basis (2 or 3)
        stack_shape = v.shape[: -self.ndim]
        v = v.reshape(-1, *v.shape[-self.ndim :])
        # Compute the adjoint
        x = self._evaluate_t(v)
        # Restore stack shape
        x = x.reshape(*stack_shape, self.count)

        # Return an ndarray
        return x

    def _evaluate_t(self, v):
        raise NotImplementedError("Subclasses should implement this")

    def mat_evaluate(self, V):
        """
        Evaluate coefficient matrix in basis

        :param V: A coefficient matrix of size `self.count`-by-
            `self.count` to be evaluated.
        :return: A multidimensional matrix of size `self.sz`-by
            -`self.sz` corresponding to the evaluation of `V` in
            this basis.
        """
        return mdim_mat_fun_conj(V, 1, len(self.sz), self._evaluate)

    def mat_evaluate_t(self, X):
        """
        Evaluate coefficient matrix in dual basis

        :param X: The coefficient array of size `self.sz`-by-`self.sz`
            to be evaluated.
        :return: The evaluation of `X` in the dual basis. This is
            `self.count`-by-`self.count`. matrix.
            If `V` is a matrix of size `self.count`-by-`self.count`,
            `B` is the change-of-basis matrix of `basis`, and `x` is a
            multidimensional matrix of size `basis.sz`-by-`basis.sz`, the
            function calculates V = B' * X * B, where the rows of `B`, rows
            of 'X', and columns of `X` are read as vectorized arrays.
        """
        return mdim_mat_fun_conj(X, len(self.sz), 1, self._evaluate_t)

    def expand(self, x):
        """
        Obtain coefficients in the basis from those in standard coordinate basis

        This is a similar function to evaluate_t but with more accuracy by using
        the cg optimizing of linear equation, Ax=b.

        :param x: An array whose last two or three dimensions are to be expanded
            the desired basis. These dimensions must equal `self.sz`.
        :return: The coefficients of `v` expanded in the desired basis.
            The last dimension of `v` is with size of `count` and the
            first dimensions of the return value correspond to
            those first dimensions of `x`.

        """
        if isinstance(x, Image) or isinstance(x, Volume):
            x = x.asnumpy()

        if x.dtype != self.dtype:
            logger.warning(
                f"{self.__class__.__name__}::expand"
                f" Inconsistent dtypes x: {x.dtype} self: {self.dtype}"
            )

        # check that last ndim values of input shape match
        # the shape of this basis
        assert (
            x.shape[-self.ndim :] == self.sz
        ), f"Last {self.ndim} dimensions of x must match {self.sz}."
        # extract number of images/volumes, or () if only one
        sz_roll = x.shape[: -self.ndim]
        # convert to standardized shape e.g. (L,L) to (1,L,L)
        x = x.reshape((-1, *self.sz))

        operator = LinearOperator(
            shape=(self.count, self.count),
            matvec=lambda v: self.evaluate_t(self.evaluate(v)),
            dtype=self.dtype,
        )

        # TODO: (from MATLAB implementation) - Check that this tolerance make sense for multiple columns in v
        tol = 10 * np.finfo(x.dtype).eps
        logger.info("Expanding array in basis")

        # number of image samples
        n_data = x.shape[0]
        v = np.zeros((n_data, self.count), dtype=self.coefficient_dtype)

        for isample in range(0, n_data):
            b = self.evaluate_t(self._cls(x[isample])).T
            # TODO: need check the initial condition x0 can improve the results or not.
            v[isample], info = cg(operator, b, tol=tol, atol=0)
            if info != 0:
                raise RuntimeError("Unable to converge!")

        # return v coefficients with the last dimension of self.count
        v = v.reshape((*sz_roll, self.count))
        return v



import logging
import os.path
from concurrent import futures
from multiprocessing import cpu_count

import mrcfile
import numpy as np

from aspire.image import Image
from aspire.operators import CTFFilter, IdentityFilter
from aspire.source import ImageSource
from aspire.utils import RelionStarFile

logger = logging.getLogger(__name__)


class RelionSource(ImageSource):
    """
    A RelionSource represents a source of picked and cropped particles stored as slices in a `.mrcs` stack.
    It must be instantiated via a STAR file, which--at a minumum--lists the particles in each `.mrcs` stack in the
    `_rlnImageName` column. The STAR file may also contain Relion-specific metadata columns. This information
    is read into dictionaries containing rows for each particle specifying its location and
    its metadata. The metadata table may be augmented or modified via helper methods found in ImageSource. It may
    store, for example, Filter objects added during preprocessing.
    """

    def __init__(
        self,
        filepath,
        data_folder=None,
        pixel_size=1,
        B=0,
        n_workers=-1,
        max_rows=None,
        memory=None,
    ):
        """
        Load STAR file at given filepath

        :param filepath: Absolute or relative path to STAR file
        :param data_folder: Path to folder w.r.t which all relative paths to .mrcs files are resolved.
            If None, the folder corresponding to filepath is used.
        :param pixel_size: the pixel size of the images in angstroms (Default 1)
        :param B: the envelope decay of the CTF in inverse square angstrom (Default 0)
        :param n_workers: Number of threads to spawn to read referenced .mrcs files (Default -1 to auto detect)
        :param max_rows: Maximum number of rows in STAR file to read. If None, all rows are read.
            Note that this refers to the max number of images to load, not the max. number of .mrcs files (which may be
            equal to or less than the number of images).
        :param memory: str or None
            The path of the base directory to use as a data store or None. If None is given, no caching is performed.
        """
        logger.info(f"Creating ImageSource from STAR file at path {filepath}")

        self.filepath = filepath
        self.data_folder = data_folder
        self.pixel_size = pixel_size
        self.B = B
        self.n_workers = n_workers
        self.max_rows = max_rows

        metadata = self.populate_metadata()

        n = len(metadata["__mrc_filepath"])
        if n == 0:
            raise RuntimeError("No mrcs files found for starfile!")

        # Peek into the first image and populate some attributes
        first_mrc_filepath = metadata["__mrc_filepath"][0]
        mrc = mrcfile.open(first_mrc_filepath)

        # Get the 'mode' (data type) - TODO: There's probably a more direct way to do this.
        mode = int(mrc.header.mode)
        dtypes = {0: "int8", 1: "int16", 2: "float32", 6: "uint16"}
        assert (
            mode in dtypes
        ), f"Only modes={list(dtypes.keys())} in MRC files are supported for now."

        dtype = dtypes[mode]

        shape = mrc.data.shape
        # the code below  accounts for the case where the first MRCS image in the STAR file has one image
        # in that case, the shape will be (resolution, resolution), whereas this code expects
        # (1, resolution, resolution). below, the shape is expanded to accomodate this expectation
        if len(shape) == 2:
            shape = (1,) + shape

        assert shape[1] == shape[2], "Only square images are supported"
        L = shape[1]
        logger.debug(f"Image size = {L}x{L}")

        # Save original image resolution that we expect to use when we start reading actual data
        self._original_resolution = L

        ImageSource.__init__(
            self, L=L, n=n, dtype=dtype, metadata=metadata, memory=memory
        )

        # CTF estimation parameters coming from Relion
        CTF_params = [
            "_rlnVoltage",
            "_rlnDefocusU",
            "_rlnDefocusV",
            "_rlnDefocusAngle",
            "_rlnSphericalAberration",
            "_rlnAmplitudeContrast",
        ]
        # If these all exist in the STAR file, we may create CTF filters for the source
        if set(CTF_params).issubset(metadata.keys()):
            # partition particles according to unique CTF parameters
            ctf_data = np.stack([metadata[k] for k in CTF_params]).T
            filter_params, filter_indices = np.unique(
                ctf_data,
                return_inverse=True,
                axis=0,
            )
            filters = []
            # for each unique CTF configuration, create a CTFFilter object
            for row in filter_params:
                filters.append(
                    CTFFilter(
                        pixel_size=self.pixel_size,
                        voltage=row[0],
                        defocus_u=row[1],
                        defocus_v=row[2],
                        defocus_ang=row[3] * np.pi / 180,  # degrees to radians
                        Cs=row[4],
                        alpha=row[5],
                        B=B,
                    )
                )
            self.unique_filters = filters
            # filter_indices stores, for each particle index, the index in
            # self.unique_filters of the filter that should be applied
            self.filter_indices = filter_indices

        # We have provided some, but not all the required params
        elif any(param in metadata for param in CTF_params):
            logger.warning(
                f"Found partially populated CTF Params."
                f"  To automatically populate CTFFilters provide {CTF_params}"
            )

        # If no CTF info in STAR, we initialize the filter values of metadata with default values
        else:
            self.unique_filters = [IdentityFilter()]
            self.filter_indices = np.zeros(self.n, dtype=int)

        logger.info(f"Populated {self.n_ctf_filters} CTFFilters from '{filepath}'")

        # Any further operations should not mutate this instance.
        self._mutable = False

    def populate_metadata(self):
        """
        Relion STAR files may contain a large number of metadata columns in addition
        to the locations of particles. We read this into a dict and add some of
        our own columns for convenience.
        """
        if self.data_folder is not None:
            if not os.path.isabs(self.data_folder):
                self.data_folder = os.path.join(
                    os.path.dirname(self.filepath), self.data_folder
                )
        else:
            self.data_folder = os.path.dirname(self.filepath)

        metadata = RelionStarFile(self.filepath).get_merged_data_block()

        # particle locations are stored as e.g. '000001@first_micrograph.mrcs'
        # in the _rlnImageName column. here, we're splitting this information
        # so we can get the particle's index in the .mrcs stack as an int
        indices_filenames = [s.split("@") for s in metadata["_rlnImageName"]]
        # __mrc_index corresponds to the integer index of the particle in the __mrc_filename stack
        # Note that this is 1-based indexing
        metadata["__mrc_index"] = np.array([int(s[0]) for s in indices_filenames])
        metadata["__mrc_filename"] = np.array([s[1] for s in indices_filenames])

        # Adding a full-filepath field to the Dataframe helps us save time later
        # Note that os.path.join works as expected when the second argument is an absolute path itself
        metadata["__mrc_filepath"] = np.array(
            [os.path.join(self.data_folder, p) for p in metadata["__mrc_filename"]]
        )

        # finally, chop off the metadata df at max_rows
        if self.max_rows is None:
            return metadata
        else:
            max_rows = min(self.max_rows, len(metadata["__mrc_filepath"]))
            return {k: v[:max_rows] for k, v in metadata.items()}

    def __str__(self):
        return f"RelionSource ({self.n} images of size {self.L}x{self.L})"

    def _images(self, indices):
        """
        Returns particle images when accessed via the `ImageSource.images` property.
        Loads particle images corresponding to `indices` from StarFile and .mrcs stacks.
        :param indices: A 1-D NumPy array of integer indices.
        :return: An `Image` object.
        """

        # check for cached images first
        if self._cached_im is not None:
            logger.debug("Loading images from cache")
            return self.generation_pipeline.forward(
                self._cached_im[indices, :, :], indices
            )

        logger.debug(f"Loading {len(indices)} images from STAR file")
        # Log the indices in case needed to debug a crash
        logger.debug(f"Indices: {indices}")

        def load_single_mrcs(filepath, indices):
            arr = mrcfile.open(filepath).data
            # if the stack only contains one image, arr will have shape (resolution, resolution)
            # the code below reshapes it to (1, resolution, resolution)
            if len(arr.shape) == 2:
                arr = arr.reshape((1,) + arr.shape)
            # __mrc_index is the 1-based index of the particle in the stack
            data = arr[self._metadata["__mrc_index"][indices] - 1, :, :]

            return indices, data

        n_workers = self.n_workers
        if n_workers < 0:
            n_workers = cpu_count() - 1

        im = np.empty(
            (len(indices), self._original_resolution, self._original_resolution),
            dtype=self.dtype,
        )

        filepaths, filepath_indices = np.unique(
            self._metadata["__mrc_filepath"], return_inverse=True
        )
        n_workers = min(n_workers, len(filepaths))

        with futures.ThreadPoolExecutor(n_workers) as executor:
            to_do = []
            for i, filepath in enumerate(filepaths):
                this_filepath_indices = np.where(filepath_indices == i)[0]
                future = executor.submit(
                    load_single_mrcs, filepath, this_filepath_indices
                )
                to_do.append(future)

            for future in futures.as_completed(to_do):
                data_indices, data = future.result()
                for idx, d in enumerate(data_indices):
                    im[np.where(indices == d)] = data[idx, :, :]

        logger.debug(f"Loading {len(indices)} images complete")

        # Finally, apply transforms to resulting Image
        return self.generation_pipeline.forward(Image(im), indices)



import logging

import finufft
import numpy as np

from aspire.nufft import Plan
from aspire.utils import complex_type

logger = logging.getLogger(__name__)


class FinufftPlan(Plan):
    def __init__(self, sz, fourier_pts, epsilon=1e-8, ntransforms=1, **kwargs):
        """
        A plan for non-uniform FFT in 2D or 3D.

        :param sz: A tuple indicating the geometry of the signal.
        :param fourier_pts: The points in Fourier space where the Fourier
            transform is to be calculated, arranged as a dimension-by-K array.
            These need to be in the range [-pi, pi] in each dimension.
        :param epsilon: The desired precision of the NUFFT.
        :param ntransforms: Optional integer indicating if you would like
            to compute a batch of `ntransforms`.
            transforms.  Implies vol_f.shape is (`ntransforms`, ...).
        """

        self.ntransforms = ntransforms

        self.sz = sz
        self.dim = len(sz)

        self.dtype = fourier_pts.dtype

        self.complex_dtype = complex_type(self.dtype)

        self.fourier_pts = np.ascontiguousarray(
            np.mod(fourier_pts + np.pi, 2 * np.pi) - np.pi
        )

        self.num_pts = fourier_pts.shape[1]

        self.epsilon = max(epsilon, np.finfo(self.dtype).eps)
        if self.epsilon != epsilon:
            logger.debug(
                f"FinufftPlan adjusted eps={self.epsilon}" f" from requested {epsilon}."
            )

        self._transform_plan = finufft.Plan(
            nufft_type=2,
            n_modes_or_dim=self.sz,
            eps=self.epsilon,
            n_trans=self.ntransforms,
            dtype=self.dtype,
        )

        self._adjoint_plan = finufft.Plan(
            nufft_type=1,
            n_modes_or_dim=self.sz,
            eps=self.epsilon,
            n_trans=self.ntransforms,
            dtype=self.dtype,
        )

        self._transform_plan.setpts(*self.fourier_pts)
        self._adjoint_plan.setpts(*self.fourier_pts)

    def transform(self, signal):
        """
        Compute the NUFFT transform using this plan instance.

        :param signal: Signal to be transformed. For a single transform,
            this should be a a 1, 2, or 3D array matching the plan `sz`.
            For a batch, signal should have shape `(ntransforms, *sz)`.

        :returns: Transformed signal of shape `num_pts` or
            `(ntransforms, num_pts)`.
        """

        sig_frame_shape = signal.shape
        # Note, there is a corner case for ntransforms == 1.
        if self.ntransforms > 1 or (
            self.ntransforms == 1 and len(signal.shape) == self.dim + 1
        ):
            assert (
                len(signal.shape) == self.dim + 1
            ), f"For multiple transforms, {self.dim}D signal should be a {self.ntransforms} element stack of {self.sz}."
            assert (
                signal.shape[0] == self.ntransforms
            ), "For multiple transforms, signal stack length should match ntransforms {self.ntransforms}."

            sig_frame_shape = signal.shape[1:]

            # finufft expects signal.ndim == dim for ntransforms = 1.
            if self.ntransforms == 1:
                signal = signal.reshape(self.sz)

        assert (
            sig_frame_shape == self.sz
        ), f"Signal frame to be transformed must have shape {self.sz}"

        # FINUFFT was designed for a complex input array
        signal = np.array(signal, copy=False, dtype=self.complex_dtype, order="C")

        result = self._transform_plan.execute(signal)

        return result

    def adjoint(self, signal):
        """
        Compute the NUFFT adjoint using this plan instance.

        :param signal: Signal to be transformed. For a single transform,
            this should be a a 1D array of len `num_pts`.
            For a batch, signal should have shape `(ntransforms, num_pts)`.

        :returns: Transformed signal `(sz)` or `(ntransforms, sz)`.
        """

        # Note, there is a corner case for ntransforms == 1.
        if self.ntransforms > 1 or (self.ntransforms == 1 and len(signal.shape) == 2):
            assert (
                len(signal.shape) == 2
            ), f"For multiple {self.dim}D adjoints, signal should be a {self.ntransforms} element stack of {self.num_pts}."
            assert (
                signal.shape[0] == self.ntransforms
            ), "For multiple transforms, signal stack length should match ntransforms {self.ntransforms}."

            # finufft is expecting flat array for 1D case.
            if self.ntransforms == 1:
                signal = signal.reshape(self.num_pts)

        # FINUFFT was designed for a complex input array
        signal = np.array(signal, copy=False, dtype=self.complex_dtype, order="C")

        result = self._adjoint_plan.execute(signal)

        return result



import importlib
import logging.config
import os
import pkgutil
from datetime import datetime
from pathlib import Path

import confuse

import aspire
from aspire.exceptions import handle_exception

# version in maj.min.bld format
__version__ = "0.11.0"


# Setup `confuse` config
config = confuse.Configuration("ASPIRE", __name__)

# Ensure the log_dir exists.
log_dir_path = Path(config["logging"]["log_dir"].get(confuse.Filename(cwd=".")))
log_dir_path.mkdir(parents=True, exist_ok=True)
# We'll reassign the evaluated log_dir back into the config so it displays well.
config["logging"]["log_dir"] = log_dir_path.as_posix()

# log output file prefix
log_prefix = config["logging"]["log_prefix"].get(str)

# DEBUG, INFO, etc.
_logging_level_names = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
console_level = config["logging"]["console_level"].as_choice(_logging_level_names)
log_file_level = config["logging"]["log_file_level"].as_choice(_logging_level_names)

# Generates file name details and opens log file defined in config file.
# The default is to use the current time stamp provided in the dictionary,
#   but that is not required if a user wishes to customize logging config.
logging.config.fileConfig(
    os.path.join(os.path.dirname(__file__), "logging.conf"),
    defaults={
        "console_level": console_level,
        "log_file_level": log_file_level,
        "log_dir": log_dir_path.as_posix(),
        "log_prefix": log_prefix,
        "dt_stamp": datetime.now().strftime("%Y-%m-%dT%H-%M-%S.%f"),
    },
)

# Log where the package resolves `config_dir()`.
logging.debug(f"ASPIRE configuration directory is {config.config_dir()}")

# Log the resolution of configuration (ie overrides).
# The list is a stack of configuration sources,
#   with each entry specifying the variables overridden.
logging.debug(
    f"ASPIRE configuration resolution details {list(aspire.config.resolve())}"
)

# Dump the config at `aspire` import time to our log.
logging.debug(f"Resolved config.yaml:\n{aspire.config.dump()}\n")

# Implements some code that writes out exceptions to 'aspire.err.log'.
if config["logging"]["log_exceptions"].get(int):
    import sys

    sys.excepthook = handle_exception


__all__ = []
for _, modname, _ in pkgutil.iter_modules(aspire.__path__):
    __all__.append(modname)  # Add module to __all_
    importlib.import_module(f"aspire.{modname}")  # Import the module



import mkl_fft
import numpy as np

from aspire.numeric.base_fft import FFT


class MKLFFT(FFT):
    """
    Define a unified wrapper class for MKL FFT functions

    To be consistent with Pyfftw, not all arguments are included.
    """

    # Note MKL does use the "workers" argument
    def fft(self, x, axis=-1, workers=-1):
        return mkl_fft.fft(
            x,
            axis=axis,
        )

    def ifft(self, x, axis=-1, workers=-1):
        return mkl_fft.ifft(
            x,
            axis=axis,
        )

    def fft2(self, x, axes=(-2, -1), workers=-1):
        return mkl_fft.fft2(
            x,
            axes=axes,
        )

    def ifft2(self, x, axes=(-2, -1), workers=-1):
        return mkl_fft.ifft2(
            x,
            axes=axes,
        )

    def fftn(self, x, axes=None, workers=-1):
        return mkl_fft.fftn(
            x,
            axes=axes,
        )

    def ifftn(self, x, axes=None, workers=-1):
        return mkl_fft.ifftn(
            x,
            axes=axes,
        )

    def fftshift(self, x, axes=None):
        # N/A in mkl_fft, use np
        return np.fft.fftshift(x, axes=axes)

    def ifftshift(self, x, axes=None):
        # N/A in mkl_fft, use np
        return np.fft.ifftshift(x, axes=axes)



import logging

import numpy as np

logger = logging.getLogger(__name__)


def fill_struct(obj=None, att_vals=None):
    """
    Fill object with attributes in a dictionary.

    If a struct is not given a new object will be created and filled.
    If the given struct has a field in att_vals, the original field will stay,
    unless specified otherwise in overwrite.
    att_vals is a dictionary with string keys, and for each key:
    if hasattr(s, key) and key in overwrite:
        pass
    else:
        setattr(s, key, att_vals[key])

    :param obj:
    :param att_vals:
    :param overwrite
    :return:
    """
    # TODO should consider making copy option - i.e that the input won't change

    if obj is None:
        obj = {}
    if att_vals is None:
        return obj

    for key in att_vals.keys():
        if key not in obj.keys():
            obj[key] = att_vals[key]

    return obj


def conj_grad(a_fun, b, cg_opt=None, init=None):
    """
    Conjugate Gradient method to solve the linear system.

    This is corresponding to the implemented version in the ASPIRE Matlab package.

    :param a_fun:  A function handle specifying the linear operation x -> Ax.
        When multiple right-hand sides are supplied, this function takes as
        input an array of shape (n, p), where n is the number of right-hand
        sides and p is the dimension of the space.
    :param b:  The vector consisting of the right hand side of Ax = b. Again,
        n different right-hand sides are given by supplying an array of shape
        (n, p).
    :param cg_opt: The parameters for the conjugate gradient method, including:
            max_iter: Maximum number of iterations (default 50).
            verbose: The extent to which information on progress should be
                output to the terminal (default 1).
            iter_callback: If non-empty, specifies a function to be called at
                the end of every iteration. In this case, iter_callback must be a
                function handle taking as single argument the info structure at
                the current iteration. For information on the info structure,
                see below (default []).
            preconditioner: If non-empty, specifies a preconditioner to be
                used in every iteration as a function handle defining the linear
                operator x -> Px (default []).
            rel_tolerance: The relative error at which to stop the algorithm,
                even if it has not yet reached the maximum number of iterations
                (default 1e-15).
            store_iterates: Defines whether to store each intermediate results
                in the info structure under the x, p and r fields. Since this
                may require a large amount of memory, this is not recommended
                (default false).
    :param init: A structure specifying the starting point of the algorithm.
            This can contain values of x or p that will be used for initialization
            (default empty).
    :return: The output result includes:
            x: The result of the conjugate gradient method after max_iter iterations
                or once the residual norm has decreased below rel_tolerance, relative.
            obj: The value of the objective function at the last iteration.
            info: A structure array containing intermediate information obtained
            during each iteration. These fields include:
            - iter: The iteration number.
            - x (for store_iterates true): The value of x.
            - r (for store_iterates true): The residual vector.
            - p (for store_iterates true): The p vector.
            - res: The square norm of the residual.
            - obj: The objective function.
    """

    def identity(input_x):
        return input_x

    default_opt = {
        "verbose": 0,
        "max_iter": 50,
        "iter_callback": [],
        "store_iterates": False,
        "rel_tolerance": 1e-15,
        "precision": b.dtype,
        "preconditioner": identity,
    }

    cg_opt = fill_struct(cg_opt, default_opt)

    default_init = {"x": None, "p": None}
    init = fill_struct(default_init, init)
    if init["x"] is None:
        x = np.zeros(b.shape, dtype=b.dtype)
    else:
        x = init["x"]

    b_norm = np.linalg.norm(b)
    r = b.copy()

    # Need the copy call to ensure that s and r are not identical in the case
    # of an identity preconditioner.
    s = cg_opt["preconditioner"](r.copy())

    if np.any(x != 0):
        if cg_opt["verbose"]:
            logger.info("[CG] Calculating initial residual")
        a_x = a_fun(x)
        r = r - a_x
        s = cg_opt["preconditioner"](r)
    else:
        a_x = np.zeros(x.shape, dtype=b.dtype)

    obj = np.real(np.sum(x.conj() * a_x, -1) - 2 * np.real(np.sum(np.conj(b * x), -1)))

    if init["p"] is None:
        p = s.copy()
    else:
        p = init["p"]

    info = fill_struct(att_vals={"iter": [0], "res": [np.linalg.norm(r)], "obj": [obj]})
    if cg_opt["store_iterates"]:
        info = fill_struct(info, att_vals={"x": [x], "r": [r], "p": [p]})

    if cg_opt["verbose"]:
        logger.info(
            "[CG] Initialized. Residual: {}. Objective: {}".format(
                np.linalg.norm(info["res"][0]), np.sum(info["obj"][0])
            )
        )

    if b_norm == 0:
        # Matlat code returns b_norm == 0, this break the Python code when b = 0
        return x, obj, info

    for i in range(1, cg_opt["max_iter"] + 1):
        if cg_opt["verbose"]:
            logger.info("[CG] Applying matrix & preconditioner")

        a_p = a_fun(p)
        old_gamma = np.real(np.sum(s.conj() * r, -1))
        alpha = old_gamma / np.real(np.sum(p.conj() * a_p, -1))
        x += alpha[..., np.newaxis] * p
        a_x += alpha[..., np.newaxis] * a_p

        r -= alpha[..., np.newaxis] * a_p
        s = cg_opt["preconditioner"](r.copy())
        new_gamma = np.real(np.sum(r.conj() * s, -1))
        beta = new_gamma / old_gamma
        p *= beta[..., np.newaxis]
        p += s

        obj = np.real(
            np.sum(x.conj() * a_x, -1) - 2 * np.real(np.sum(np.conj(b * x), -1))
        )
        res = np.sqrt(np.sum(r**2, -1))
        info["iter"].append(i)
        info["res"].append(res)
        info["obj"].append(obj)
        if cg_opt["store_iterates"]:
            info["x"].append(x)
            info["r"].append(r)
            info["p"].append(p)

        if cg_opt["verbose"]:
            logger.info(
                "[CG] Iteration {}. Residual: {}. Objective: {}".format(
                    i, np.linalg.norm(info["res"][i]), np.sum(info["obj"][i])
                )
            )

        if np.all(res < b_norm * cg_opt["rel_tolerance"]):
            break

    if i == cg_opt["max_iter"]:
        logger.warning("[CG] Conjugate gradient reached maximum number of iterations!")

    return x, obj, info



import functools

import click

# Wraps click.option to set global defaults for ASPIRE commands.
# Currently this just sets show_default.
# To use, also import the following when importing and defining click commands:
# # `import aspire.commands._config`
click.option = functools.partial(click.option, show_default=True)



import logging

from aspire.source import ImageSource

logger = logging.getLogger(__name__)


class OrientEstSource(ImageSource):
    """
    Derived an ImageSource class for updating orientation information
    """

    def __init__(self, src, orient_method):
        """
        Initialize an Orientation ImageSource object from original ImageSource

        :param src: Original ImageSource object after 2D classification
        :param orient_method: object specifying orientation estimation method
        """

        super().__init__(src.L, src.n, dtype=src.dtype, metadata=src._metadata.copy())
        self._im = None
        self.orient_method = orient_method
        self.rotations = orient_method.rotations

    def _images(self, indices):
        return self._im[indices, :, :]



import logging

import numpy as np

from aspire.utils import Rotation

logger = logging.getLogger(__name__)


class SyncVotingMixin(object):
    """
    SyncVotingMixin is a mixin implementing methods for the synchronization voting algorithm
    which are shared by CLSynVoting and CLSymmetryC3C4
    """

    def _rotratio_eulerangle_vec(self, clmatrix, i, j, good_k, n_theta):
        """
        Compute the rotation that takes image i to image j

        Given a common lines matrix, where the index of each common line
        is in the range of n_theta and a list of good image k from voting results.
        :param clmatrix: The common lines matrix
        :param i: The i image
        :param j: The j image
        :param good_k: The list of good images k from voting algorithm
        :param n_theta: The number of points in the theta direction (common lines)
        :return: The rotation matrix that takes image i to image j for good index of k.
        """

        if i == j:
            return []

        # Prepare the theta values from the differences of common line indices
        # C1, C2, and C3 are unit circles of image i, j, and k
        # cl_diff1 is for the angle on C1 created by its intersection with C3 and C2.
        # cl_diff2 is for the angle on C2 created by its intersection with C1 and C3.
        # cl_diff3 is for the angle on C3 created by its intersection with C2 and C1.
        cl_diff1 = clmatrix[i, good_k] - clmatrix[i, j]  # for theta1
        cl_diff2 = clmatrix[j, good_k] - clmatrix[j, i]  # for - theta2
        cl_diff3 = clmatrix[good_k, j] - clmatrix[good_k, i]  # for theta3

        # Calculate the cos values of rotation angles between i an j images for good k images
        c_alpha, good_idx = self._get_cos_phis(cl_diff1, cl_diff2, cl_diff3, n_theta)
        if len(c_alpha) == 0:
            return None
        alpha = np.arccos(c_alpha)

        # Convert the Euler angles with ZYZ conversion to rotation matrices
        angles = np.zeros((alpha.shape[0], 3))
        angles[:, 0] = clmatrix[i, j] * 2 * np.pi / n_theta + np.pi / 2
        angles[:, 1] = alpha
        angles[:, 2] = -np.pi / 2 - clmatrix[j, i] * 2 * np.pi / n_theta
        r = Rotation.from_euler(angles).matrices

        return r[good_idx, :, :]

    def _vote_ij(self, clmatrix, n_theta, i, j, k_list):
        """
        Apply the voting algorithm for images i and j.

        clmatrix is the common lines matrix, constructed using angular resolution,
        n_theta. k_list are the images to be used for voting of the pair of images
        (i ,j).
        :param clmatrix: The common lines matrix
        :param n_theta: The number of points in the theta direction (common lines)
        :param i: The i image
        :param j: The j image
        :param k_list: The list of images for the third image for voting algorithm
        :return:  good_k, the list of all third images in the peak of the histogram
            corresponding to the pair of images (i,j)
        """

        if i == j or clmatrix[i, j] == -1:
            return []

        # Some of the entries in clmatrix may be zero if we cleared
        # them due to small correlation, or if for each image
        # we compute intersections with only some of the other images.
        #
        # Note that as long as the diagonal of the common lines matrix is
        # -1, the conditions (i != j) && (j != k) are not needed, since
        # if i == j then clmatrix[i, k] == -1 and similarly for i == k or
        # j == k. Thus, the previous voting code (from the JSB paper) is
        # correct even though it seems that we should test also that
        # (i != j) && (i != k) && (j != k), and only (i != j) && (i != k)
        #  as tested there.
        cl_idx12 = clmatrix[i, j]
        cl_idx21 = clmatrix[j, i]
        k_list = k_list[
            (k_list != i) & (clmatrix[i, k_list] != -1) & (clmatrix[j, k_list] != -1)
        ]
        cl_idx13 = clmatrix[i, k_list]
        cl_idx31 = clmatrix[k_list, i]
        cl_idx23 = clmatrix[j, k_list]
        cl_idx32 = clmatrix[k_list, j]

        # Prepare the theta values from the differences of common line indices
        # C1, C2, and C3 are unit circles of image i, j, and k
        # cl_diff1 is for the angle on C1 created by its intersection with C3 and C2.
        # cl_diff2 is for the angle on C2 created by its intersection with C1 and C3.
        # cl_diff3 is for the angle on C3 created by its intersection with C2 and C1.
        cl_diff1 = cl_idx13 - cl_idx12
        cl_diff2 = cl_idx21 - cl_idx23
        cl_diff3 = cl_idx32 - cl_idx31
        # Calculate the cos values of rotation angles between i an j images for good k images
        cos_phi2, good_idx = self._get_cos_phis(cl_diff1, cl_diff2, cl_diff3, n_theta)

        if np.any(np.abs(cos_phi2) - 1 > 1e-12):
            logger.warning(
                f"Globally Consistent Angular Reconstruction (GCAR) exists"
                f" numerical problem: abs(cos_phi2) > 1, with the"
                f" difference of {np.abs(cos_phi2)-1}."
            )
        cos_phi2 = np.clip(cos_phi2, -1, 1)

        # Store angles between i and j induced by each third image k.
        phis = cos_phi2
        # Sore good indices of l in k_list of the image that creates that angle.
        inds = k_list[good_idx]

        if phis.shape[0] == 0:
            return []

        # Parameters used to compute the smoothed angle histogram.
        ntics = 60
        angles_grid = np.linspace(0, 180, ntics, True)
        # Get angles between images i and j for computing the histogram
        angles = np.arccos(phis[:]) * 180 / np.pi
        # Angles that are up to 10 degrees apart are considered
        # similar. This sigma ensures that the width of the density
        # estimation kernel is roughly 10 degrees. For 15 degrees, the
        # value of the kernel is negligible.
        sigma = 3.0

        # Compute the histogram of the angles between images i and j
        squared_values = np.add.outer(np.square(angles), np.square(angles_grid))
        angles_hist = np.sum(
            np.exp(
                (2 * np.multiply.outer(angles, angles_grid) - squared_values)
                / (2 * sigma**2)
            ),
            0,
        )

        # We assume that at the location of the peak we get the true angle
        # between images i and j. Find all third images k, that induce an
        # angle between i and j that is at most 10 off the true angle.
        # Even for debugging, don't put a value that is smaller than two
        # tics, since the peak might move a little bit due to wrong k images
        # that accidentally fall near the peak.
        peak_idx = angles_hist.argmax()
        idx = np.abs(angles - angles_grid[peak_idx]) < 360 / ntics
        good_k = inds[idx]
        return good_k.astype("int")

    def _get_cos_phis(self, cl_diff1, cl_diff2, cl_diff3, n_theta):
        """
        Calculate cos values of rotation angles between i and j images

        Given C1, C2, and C3 are unit circles of image i, j, and k, compute
        resulting cos values of rotation angles between i an j images when both
        of them are intersecting with k.

        To ensure that the smallest singular value is big enough, controlled by
        the determinant of the matrix,
           C=[  1  c1  c2 ;
               c1   1  c3 ;
               c2  c3   1 ],
        we therefore use the condition below
               1+2*c1*c2*c3-(c1^2+c2^2+c3^2) > 1.0e-5,
        so the matrix is far from singular.

        :param cl_diff1: Difference of common line indices on C1 created by
            its intersection with C3 and C2
        :param cl_diff2: Difference of common line indices on C2 created by
            its intersection with C1 and C3
        :param cl_diff3: Difference of common line indices on C3 created by
            its intersection with C2 and C1
        :param n_theta: The number of points in the theta direction (common lines)
        :return: cos values of rotation angles between i and j images
            and indices for good k
        """

        # Calculate the theta values from the differences of common line indices
        # C1, C2, and C3 are unit circles of image i, j, and k
        # theta1 is the angle on C1 created by its intersection with C3 and C2.
        # theta2 is the angle on C2 created by its intersection with C1 and C3.
        # theta3 is the angle on C3 created by its intersection with C2 and C1.
        theta1 = cl_diff1 * 2 * np.pi / n_theta
        theta2 = cl_diff2 * 2 * np.pi / n_theta
        theta3 = cl_diff3 * 2 * np.pi / n_theta

        c1 = np.cos(theta1)
        c2 = np.cos(theta2)
        c3 = np.cos(theta3)

        # Each common-line corresponds to a point on the unit sphere. Denote the
        # coordinates of these points by (Pix, Piy Piz), and put them in the matrix
        #   M=[ P1x  P2x  P3x ;
        #       P1y  P2y  P3y ;
        #       P1z  P2z  P3z ].
        #
        # Then the matrix
        #   C=[  1  c1  c2 ;
        #       c1   1  c3 ;
        #       c2  c3   1 ],
        # where c1, c2, c3 are given above, is given by C = M.T @ M.
        # For the points P1, P2, and P3 to form a triangle on the unit sphere, a
        # necessary and sufficient condition is for C to be positive definite. This
        # is equivalent to
        #       1+2*c1*c2*c3-(c1^2+c2^2+c3^2) > 0.
        # However, this may result in a triangle that is too flat, that is, the
        # angle between the projections is very close to zero. We therefore use the
        # condition below
        #       1+2*c1*c2*c3-(c1^2+c2^2+c3^2) > 1.0e-5.
        # This ensures that the smallest singular value (which is actually
        # controlled by the determinant of C) is big enough, so the matrix is far
        # from singular. This condition is equivalent to computing the singular
        # values of C, followed by checking that the smallest one is big enough.

        cond = 1 + 2 * c1 * c2 * c3 - (np.square(c1) + np.square(c2) + np.square(c3))
        good_idx = np.nonzero(cond > 1e-5)[0]

        # Calculated cos values of angle between i and j images
        cos_phi2 = (c3[good_idx] - c1[good_idx] * c2[good_idx]) / (
            np.sin(theta1[good_idx]) * np.sin(theta2[good_idx])
        )
        return cos_phi2, good_idx



class FFT:
    """
    Define a customized interface for FFT functions

    To make consistent among Pyfftw, Scipyfft and cupy fft,
    not all arguments are included.
    """

    def fft(self, x, axis=-1, workers=-1):
        raise NotImplementedError("subclasses must implement this")

    def ifft(self, x, axis=-1, workers=-1):
        raise NotImplementedError("subclasses must implement this")

    def fft2(self, x, axes=(-2, -1), workers=-1):
        raise NotImplementedError("subclasses must implement this")

    def ifft2(self, x, axes=(-2, -1), workers=-1):
        raise NotImplementedError("subclasses must implement this")

    def fftn(self, x, axes=None, workers=-1):
        raise NotImplementedError("subclasses must implement this")

    def ifftn(self, x, axes=None, workers=-1):
        raise NotImplementedError("subclasses must implement this")

    def fftshift(self, x, axes=None):
        raise NotImplementedError("subclasses must implement this")

    def ifftshift(self, x, axes=None):
        raise NotImplementedError("subclasses must implement this")

    def centered_ifft(self, x, axis=-1, workers=-1):
        x = self.ifftshift(x, axes=axis)
        x = self.ifft(x, axis=axis, workers=workers)
        x = self.fftshift(x, axes=axis)
        return x

    def centered_fft(self, x, axis=-1, workers=-1):
        x = self.ifftshift(x, axes=axis)
        x = self.fft(x, axis=axis, workers=workers)
        x = self.fftshift(x, axes=axis)
        return x

    def centered_ifft2(self, x, axes=(-2, -1), workers=-1):
        x = self.ifftshift(x, axes=axes)
        x = self.ifft2(x, axes=axes, workers=workers)
        x = self.fftshift(x, axes=axes)
        return x

    def centered_fft2(self, x, axes=(-2, -1), workers=-1):
        x = self.ifftshift(x, axes=axes)
        x = self.fft2(x, axes=axes, workers=workers)
        x = self.fftshift(x, axes=axes)
        return x

    def centered_ifftn(self, x, axes=None, workers=-1):
        x = self.ifftshift(x, axes=axes)
        x = self.ifftn(x, axes=axes, workers=workers)
        x = self.fftshift(x, axes=axes)
        return x

    def centered_fftn(self, x, axes=None, workers=-1):
        x = self.ifftshift(x, axes=axes)
        x = self.fftn(x, axes=axes, workers=workers)
        x = self.fftshift(x, axes=axes)
        return x

    def mdim_ifftshift(self, x, dims=None):
        """
        Multi-dimensional FFT unshift

        :param x: The array to be unshifted.
        :param dims: An array of dimension indices along which the unshift should occur.
            If None, the unshift is performed along all dimensions.
        :return: The x array unshifted along the desired dimensions.
        """
        if dims is None:
            dims = range(0, x.ndim)
        for dim in dims:
            x = self.ifftshift(x, dim)
        return x

    def mdim_fftshift(self, x, dims=None):
        """
        Multi-dimensional FFT shift

        :param x: The array to be shifted.
        :param dims: An array of dimension indices along which the shift should occur.
            If None, the shift is performed along all dimensions.
        :return: The x array shifted along the desired dimensions.
        """
        if dims is None:
            dims = range(0, x.ndim)
        for dim in dims:
            x = self.fftshift(x, dim)
        return x



import logging

import numpy as np

from aspire.nufft import anufft
from aspire.numeric import fft
from aspire.operators import evaluate_src_filters_on_grid
from aspire.reconstruction import Estimator, FourierKernel
from aspire.volume import rotated_grids

logger = logging.getLogger(__name__)


class MeanEstimator(Estimator):
    def compute_kernel(self):
        _2L = 2 * self.src.L
        kernel = np.zeros((_2L, _2L, _2L), dtype=self.dtype)
        sq_filters_f = np.square(evaluate_src_filters_on_grid(self.src))

        for i in range(0, self.src.n, self.batch_size):
            _range = np.arange(i, min(self.src.n, i + self.batch_size), dtype=int)
            pts_rot = rotated_grids(self.src.L, self.src.rotations[_range, :, :])
            weights = sq_filters_f[:, :, _range]
            weights *= self.src.amplitudes[_range] ** 2

            if self.src.L % 2 == 0:
                weights[0, :, :] = 0
                weights[:, 0, :] = 0

            pts_rot = pts_rot.reshape((3, -1))
            weights = np.transpose(weights, (2, 0, 1)).flatten()

            kernel += (
                1
                / (self.src.n * self.src.L**4)
                * anufft(weights, pts_rot[::-1], (_2L, _2L, _2L), real=True)
            )

        # Ensure symmetric kernel
        kernel[0, :, :] = 0
        kernel[:, 0, :] = 0
        kernel[:, :, 0] = 0

        logger.info("Computing non-centered Fourier Transform")
        kernel = fft.mdim_ifftshift(kernel, range(0, 3))
        kernel_f = fft.fft2(kernel, axes=(0, 1, 2))
        kernel_f = np.real(kernel_f)

        return FourierKernel(kernel_f, centered=False)



import logging

logger = logging.getLogger(__name__)


class Denoiser:
    """
    Define a base class for denoising 2D images
    """

    def __init__(self, src):
        """
        Initialize an object for denoising 2D images from the image source

        :param src: The source object of 2D images with metadata
        """
        self.src = src
        self.dtype = src.dtype
        self.nimg = src.n

    def denoise(self):
        """
        Precompute for Denoiser and DenoisedImageSource for 2D images
        """
        raise NotImplementedError("subclasses must implement this")

    def image(self, istart=0, batch_size=512):
        """
        Obtain a batch size of 2D images after denosing by a specified method
        """
        raise NotImplementedError("subclasses must implement this")



import logging

import numpy as np

logger = logging.getLogger(__name__)


class Cell2D:
    """
    Define a base class of Cell to perform similar cell functions in MATLAB.

    """

    def __init__(self, rows, cols, dtype=np.float32):
        self.dtype = np.dtype(dtype)
        self.dtype = dtype
        self.rows = rows
        self.cols = cols
        self.nrow = np.size(rows)
        self.ncol = np.size(cols)

        size = 0
        for i in range(0, np.size(rows)):
            for j in range(0, np.size(cols)):
                size += rows[i] * cols[j]
        self.size = size
        self.cell_list = []
        for i in range(0, np.size(rows)):
            for j in range(0, np.size(cols)):
                self.cell_list.append(np.zeros((rows[i], cols[j]), dtype=self.dtype))

    def mat2cell(self, mat, rows, cols):
        if np.sum(rows) != np.size(mat, 0) or np.sum(cols) != np.size(mat, 1):
            raise RuntimeError(
                "Matrix is not compatible wth the cell array in rows or cols size!"
            )
        offset = 0
        offsetr = 0
        for i in range(0, self.nrow):
            offsetc = 0
            for j in range(0, self.ncol):
                self.cell_list[offset][:] = mat[
                    offsetr : offsetr + rows[i], offsetc : offsetc + cols[j]
                ]
                offset += 1
                offsetc += cols[j]
            offsetr += rows[i]
        return self.cell_list



"""
Define a BlkDiagMatrix module which implements operations for
block diagonal matrices as used by ASPIRE.
"""

import numpy as np
from numpy.linalg import norm, solve
from scipy.linalg import block_diag

from aspire.utils import make_psd
from aspire.utils.cell import Cell2D


class BlkDiagMatrix:
    """
    Define a BlkDiagMatrix class which implements operations for
    block diagonal matrices as used by ASPIRE.

    Currently BlkDiagMatrix is implemented only for square blocks.
    While in the future this can be extended, at this time assigning
    a non square array will raise NotImplementedError.
    """

    # Developers' Note:
    # All instances of this class should have priority over ndarray ops
    #   because we implement them here ourselves.
    # This is a more np current implementation of __array_priority__
    #   operator precedence schedule.
    # Mainly this effects rmul, radd, rsub eg:
    #   blk_y = scalar_a + ( scalar_b * blk_x)
    __array_ufunc__ = None

    def __init__(self, partition, dtype=np.float32):
        """
        Instantiate a BlkDiagMatrix.

        :param partition: The matrix block partition
            in the form of a `nblock`-element list storing all shapes of
            diagonal matrix blocks, where `partition[i]` corresponds to
            the shape (number of rows and columns) of the `i` matrix block.
        :param dtype: Datatype for blocks, defaults to np.float32.
        :return: BlkDiagMatrix instance.
        """

        self.nblocks = len(partition)
        self.dtype = np.dtype(dtype)
        self.data = [None] * self.nblocks
        self._cached_blk_sizes = np.array(partition)
        if len(partition):
            assert self._cached_blk_sizes.shape[1] == 2

    def reset_cache(self):
        """
        Resets this objects internal cache. This should trigger the cache
        to be recalculated on the next request (ie lazily).
        """

        self._cached_blk_sizes = None

    def append(self, blk):
        """
        Append `blk` to `self`. Used to incrementally build up a BlkDiagMatrix
        instance where the number of blocks and/or shapes are derived
        incrementally.

        :param blk: Block to append (ndarray).
        """

        self.data.append(blk)
        self.nblocks += 1
        self.reset_cache()

    def __repr__(self):
        """
        String represention describing instance.
        """
        return "BlkDiagMatrix({}, {})".format(repr(self.nblocks), repr(self.dtype))

    def copy(self):
        """
        Returns new BlkDiagMatrix which is a copy of `self`.

        :return BlkDiagMatrix like self
        """

        C = BlkDiagMatrix(self.partition, dtype=self.dtype)

        for i in range(self.nblocks):
            C[i] = self[i].copy()

        return C

    # Manually overload list methods,
    #   This is just for syntax which allows us to reference self[i] etc
    #     instead of writing self.data all the time. You may use either.
    def __getitem__(self, key):
        """
        Convenience wrapper, getter on self.data.
        """

        return self.data[key]

    def __setitem__(self, key, value):
        """
        Convenience wrapper, setter on self.data.
        """

        self.data[key] = value
        self.reset_cache()

    def __len__(self):
        """
        Convenience function for getting nblocks.
        """

        return self.nblocks

    def _is_scalar_type(self, x):
        """
        Internal helper function checking scalar-ness for elementwise ops.

        Essentially we are checking for a single numeric object, as opposed to
        something like an `ndarray` or `BlkDiagMatrix`. We do this by
        checking `numpy.isscalar(x)`.

        In the future this check may require extension to include ASPIRE or
        other third party types beyond what is provided by numpy, so we
        implement it now as a class method.

        :param x: Value to check

        :return: bool.
        """

        return np.isscalar(x)

    def __check_size_compatible_add(self, other):
        """
        Sanity check two BlkDiagMatrix instances are compatible in size
        for addition operators. (Same size)

        :param other: The BlkDiagMatrix to compare with self.
        """

        if np.any(self.partition != other.partition):
            # be helpful and find the first one as an example
            for _i, (a, b) in enumerate(zip(self.partition, other.partition)):
                if any(a != b):
                    break
            raise RuntimeError(
                "Block i={} of BlkDiagMatrix instances are "
                "not same shape {} {}".format(_i, a, b)
            )

    def __check_size_compatible_mul(self, other):
        """
        Sanity check two BlkDiagMatrix instances are compatible in size
        for multiplication operators. (m n) @ (n k).

        :param other: The BlkDiagMatrix to compare with self.
        """

        for _i, a in enumerate(self.partition):
            b = other.partition[_i]
            if a[1] != b[0]:
                raise RuntimeError(
                    "Block i={} of BlkDiagMatrix instances are "
                    "not compatible. {} {}".format(_i, a, b)
                )

    def __check_dtype_compatible(self, other):
        """
        Sanity check two BlkDiagMatrix instances are compatible in dtype.

        :param other: The BlkDiagMatrix to compare with self.
        """

        if self.dtype != other.dtype:
            raise RuntimeError(
                "BlkDiagMatrix received different types,"
                "self: {} and other: {}.  Please validate and cast"
                " as appropriate.".format(self.dtype, other.dtype)
            )

    def __check_compatible(self, other, size_compat="add"):
        """
        Sanity check two BlkDiagMatrix instances are compatible in size.

        :param other: The BlkDiagMatrix to compare with self.
        """

        if not isinstance(other, BlkDiagMatrix):
            raise NotImplementedError(
                "Currently BlkDiagMatrix only interfaces "
                "with its own instances, got {}".format(repr(other))
            )

        if len(self) != len(other):
            raise RuntimeError(
                "Number of blocks {} {} are not equal.".format(len(self), len(other))
            )

        if size_compat == "add":
            self.__check_size_compatible_add(other)
        elif size_compat == "mul":
            self.__check_size_compatible_mul(other)
        else:
            raise RuntimeError("Unknown compatibility type {}".format(size_compat))

        self.__check_dtype_compatible(other)

    @property
    def is_square(self):
        """
        Check if all blocks are square.

        :return: boolean
        """
        return all([shp[0] == shp[1] for shp in self.partition])

    @property
    def isfinite(self):
        """
        Check if all blocks in diag matrix are finite.

        Calls numpy.isfinite for every entry in self.  This has the effect of
        checking values are not += `inf` or `nan`s.

        :return: Bool.
        """

        for blk in self:
            if not np.all(np.isfinite(blk)):
                return False
        else:
            return True

    def add(self, other, inplace=False):
        """
        Define the elementwise addition of BlkDiagMatrix instance.

        :param other: The rhs BlkDiagMatrix instance.
        :param inplace: Boolean, when set to True change values in place,
            otherwise return a new instance (default).
        :return:  BlkDiagMatrix instance with elementwise sum equal
            to self + other.
        """

        if self._is_scalar_type(other):
            return self.__scalar_add(other, inplace=inplace)

        self.__check_compatible(other)

        if inplace:
            for i in range(self.nblocks):
                self[i] += other[i]

            C = self
        else:
            C = BlkDiagMatrix(self.partition, dtype=self.dtype)

            for i in range(self.nblocks):
                C[i] = self[i] + other[i]

        return C

    def __add__(self, other):
        """
        Operator overloading for addition.
        """

        return self.add(other)

    def __iadd__(self, other):
        """
        Operator overloading for in-place addition.
        """

        return self.add(other, inplace=True)

    def __radd__(self, other):
        """
        Convenience function for elementwise scalar addition.
        """

        return self.add(other)

    def __scalar_add(self, scalar, inplace=False):
        """
        Define the element addition of BlkDiagMatrix instance.

        :param scalar: constant addend value.
        :param inplace: Boolean, when set to True change values in place,
            otherwise return a new instance (default).
        :return:  BlkDiagMatrix instance with elementwise sum equal
            to self + other.
        """

        assert self._is_scalar_type(scalar)

        if inplace:
            C = self
        else:
            C = self.copy()

        for i in range(self.nblocks):
            C[i] += scalar

        return C

    def sub(self, other, inplace=False):
        """
        Define the element subtraction of BlkDiagMatrix instance.

        :param other: The rhs BlkDiagMatrix instance.
        :param inplace: Boolean, when set to True change values in place,
            otherwise return a new instance (default).
        :return: A BlkDiagMatrix instance with elementwise subraction equal to
            self - other.
        """

        if self._is_scalar_type(other):
            return self.__scalar_sub(other, inplace=inplace)

        self.__check_compatible(other)

        if inplace:
            for i in range(self.nblocks):
                self[i] -= other[i]

            C = self
        else:
            C = BlkDiagMatrix(self.partition, dtype=self.dtype)

            for i in range(self.nblocks):
                C[i] = self[i] - other[i]

        return C

    def __sub__(self, other):
        """
        Operator overloading for subtraction.
        """

        return self.sub(other)

    def __isub__(self, other):
        """
        Operator overloading for in-place subtraction.
        """

        if self._is_scalar_type(other):
            return self.__scalar_sub(other, inplace=True)

        return self.sub(other, inplace=True)

    def __rsub__(self, other):
        """
        Convenience function for elementwise scalar subtraction.
        """

        # Note, the case of BlkDiagMatrix_L - BlkDiagMatrix_R would be
        #   evaluated as L.sub(R), so this is only for other
        #   Object - BlkDiagMatrix situations, namely scalars.

        return -(self - other)

    def __scalar_sub(self, scalar, inplace=False):
        """
        Define the elementwise subtraction from BlkDiagMatrix instance.

        :param scalar: constant subtractend value.
        :param inplace: bool, when false (default) return new instance.
        :return:  BlkDiagMatrix instance with elementwise sum equal to
            self + other.
        """

        assert self._is_scalar_type(scalar)

        if inplace:
            C = self
        else:
            C = self.copy()

        for i in range(self.nblocks):
            C[i] -= scalar

        return C

    def matmul(self, other, inplace=False):
        """
        Compute the matrix multiplication of two BlkDiagMatrix instances.

        :param other: The rhs BlkDiagMatrix instance.
        :param inplace: Boolean, when set to True change values in place,
            otherwise return a new instance (default).
        :return: A BlkDiagMatrix of self @ other.
        """

        if not isinstance(other, BlkDiagMatrix):
            if inplace:
                raise RuntimeError(
                    "`inplace` method not supported when "
                    "mixing `BlkDiagMatrix` and `Numpy`."
                )
            return self.apply(other)

        self.__check_compatible(other, size_compat="mul")

        if inplace:
            for i in range(self.nblocks):
                self[i] @= other[i]
            C = self
        else:
            C = BlkDiagMatrix(self.partition, dtype=self.dtype)

            for i in range(self.nblocks):
                C[i] = self[i] @ other[i]

        return C

    def __matmul__(self, other):
        """
        Operator overload for matrix multiply of BlkDiagMatrix instances.
        """

        return self.matmul(other)

    def __rmatmul__(self, lhs):
        """
        Compute the right matrix multiplication with a BlkDiagMatrix instance,
        and a numpy array, lhs @ self.

        :param other: The lhs Numpy instance.
        :return: Returns numpy array representing `other @ self`.
        """

        # Note, we should only hit this method when mixing BlkDiagMatrix with numpy.
        #   This is because if both a and b are BlkDiagMatrix,
        #   then a@b would be handled first by a.__matmul__(b), never reaching here.
        if not isinstance(lhs, np.ndarray):
            raise RuntimeError(
                "__rmatmul__ only defined for np.ndarray @ BlkDiagMatrix."
            )

        return self.rapply(lhs)

    def __imatmul__(self, other):
        """
        Operator overload for in-place matrix multiply of BlkDiagMatrix
         instances.
        """

        return self.matmul(other, inplace=True)

    def mul(self, val, inplace=False):
        """
        Compute the numeric multiplication of a BlkDiagMatrix instance and a
        scalar.

        :param other: The rhs BlkDiagMatrix instance.
        :param inplace: Boolean, when set to True change values in place,
            otherwise return a new instance (default).
        :return: A BlkDiagMatrix of self * other.
        """

        if isinstance(val, BlkDiagMatrix):
            raise RuntimeError(
                "Attempt numeric multiplication (*,mul) of two "
                "BlkDiagMatrix instances, try (matmul,@)."
            )

        elif not self._is_scalar_type(val):
            raise RuntimeError(
                "Attempt numeric multiplication (*,mul) of a "
                "BlkDiagMatrix and {}.".format(type(val))
            )

        if inplace:
            for i in range(self.nblocks):
                self[i] *= val

            C = self
        else:
            C = BlkDiagMatrix(self.partition, dtype=self.dtype)

            for i in range(self.nblocks):
                C[i] = self[i] * val

        return C

    def __mul__(self, val):
        """
        Operator overload for BlkDiagMatrix scalar multiply.
        """

        return self.mul(val)

    def __imul__(self, val):
        """
        Operator overload for in-place BlkDiagMatrix scalar multiply.
        """

        return self.mul(val, inplace=True)

    def __rmul__(self, other):
        """
        Convenience function, elementwise rmul commutes to mul.
        """

        return self.mul(other)

    def neg(self):
        """
        Compute the unary negation of BlkDiagMatrix instance.

        :return: A BlkDiagMatrix like self.
        """

        C = BlkDiagMatrix(self.partition, dtype=self.dtype)

        for i in range(self.nblocks):
            C[i] = -self[i]

        return C

    def __neg__(self):
        """
        Operator overload for unary negation of BlkDiagMatrix instance.
        """

        return self.neg()

    def abs(self):
        """
        Compute the elementwise absolute value of BlkDiagMatrix instance.

        :return: A BlkDiagMatrix like self.
        """

        C = BlkDiagMatrix(self.partition, dtype=self.dtype)

        for i in range(self.nblocks):
            C[i] = np.abs(self[i])

        return C

    def __abs__(self):
        """
        Operator overload for absolute value of BlkDiagMatrix instance.
        """

        return self.abs()

    def pow(self, val, inplace=False):
        """
        Compute the elementwise power of BlkDiagMatrix instance.

        :param inplace: Boolean, when set to True change values in place,
            otherwise return a new instance (default).
        :return: A BlkDiagMatrix like self.
        """

        if inplace:
            for i in range(self.nblocks):
                self[i] **= val
            C = self
        else:
            C = BlkDiagMatrix(self.partition, dtype=self.dtype)
            for i in range(self.nblocks):
                C[i] = np.power(self[i], val)
        return C

    def __pow__(self, val):
        """
        Operator overload for inplace pow of BlkDiagMatrix instance.
        """

        return self.pow(val)

    def __ipow__(self, val):
        """
        Compute the in-place elementwise power of BlkDiagMatrix instance.

        :return: self raised to power, elementwise.
        """

        return self.pow(val, inplace=True)

    def norm(self):
        """
        Compute the norm of a BlkDiagMatrix instance.

        :param inplace: Boolean, when set to True change values in place,
            otherwise return a new instance (default).
        :return: The norm of the BlkDiagMatrix instance.
        """

        return np.max([norm(blk, ord=2) for blk in self])

    def transpose(self):
        """
        Get the transpose matrix of a BlkDiagMatrix instance.

        :return: The corresponding transpose form as a BlkDiagMatrix.
        """

        T = BlkDiagMatrix(self.partition, dtype=self.dtype)

        for i in range(self.nblocks):
            T[i] = self[i].T

        return T

    @property
    def T(self):
        """
        Syntactic sugar for self.transpose().
        """

        return self.transpose()

    def dense(self):
        """
        Convert list representation of BlkDiagMatrix instance into full matrix.

        :param blk_diag: The BlkDiagMatrix instance.
        :return: The BlkDiagMatrix instance including the zero elements of
        non-diagonal blocks.
        """

        return block_diag(*self.data)

    @property
    def partition(self):
        """
        Return the partitions (block sizes) of this BlkDiagMatrix

        :return: The matrix block partition in the form of a
        K-element list storing all shapes of K diagonal matrix blocks,
        where `partition[i]` corresponds to the shape (number of rows and
        columns) of the `i` diagonal matrix block.
        """

        if self._cached_blk_sizes is None:
            blk_sizes = np.empty((self.nblocks, 2), dtype=int)
            for i, blk in enumerate(self.data):
                blk_sizes[i] = np.shape(blk)
            self._cached_blk_sizes = blk_sizes
        return self._cached_blk_sizes

    def solve(self, Y):
        """
        Solve a linear system involving a block diagonal matrix.

        :param Y: The right-hand side in the linear system.  May be a matrix
            consisting of coefficient vectors, in which case each column is
            solved for separately.

        :return: The result of solving the linear system formed by the matrix.
        """

        rows = self.partition[:, 0]
        if sum(rows) != Y.shape[0]:
            raise RuntimeError("Sizes of `self` and `Y` are not compatible.")

        # Use `np.linalg.solve` for square matrices/blocks.
        #   If user requires solving non square, we'll need to extend for
        #   lstsq or qr,triangle solvers.
        if not self.is_square:
            raise NotImplementedError(
                "BlkDiagMatrix.solve is only defined for square arrays. "
                "If you require solving non square BlkDiagMatrix please "
                "report to developers."
            )

        vector = False
        if np.ndim(Y) == 1:
            Y = Y[:, np.newaxis]
            vector = True

        cols = np.array([np.size(Y, 1)])
        cellarray = Cell2D(rows, cols, dtype=Y.dtype)
        Y = cellarray.mat2cell(Y, rows, cols)
        X = []
        for i in range(0, self.nblocks):
            X.append(solve(self[i], Y[i]))
        X = np.concatenate(X, axis=0)

        if vector:
            X = X[:, 0]

        return X

    def apply(self, X):
        """
        Define the apply option of a block diagonal matrix with a matrix of
        coefficient vectors.

        :param X: Coefficient matrix, each column is a coefficient vector.
        :return: A matrix with new coefficient vectors.
        """

        cols = self.partition[:, 1]

        if np.sum(cols) != np.size(X, 0):
            raise RuntimeError("Sizes of matrix `self` and `X` are not compatible.")

        vector = False
        if np.ndim(X) == 1:
            X = X[:, np.newaxis]
            vector = True

        rows = np.array(
            [
                np.size(X, 1),
            ]
        )
        cellarray = Cell2D(cols, rows, dtype=X.dtype)
        x_cell = cellarray.mat2cell(X, cols, rows)
        Y = []
        for i in range(0, self.nblocks):
            mat = self[i] @ x_cell[i]
            Y.append(mat)
        Y = np.concatenate(Y, axis=0)

        if vector:
            Y = Y[:, 0]

        return Y

    def rapply(self, X):
        """
        Right apply.  Given a matrix of coefficient vectors,
        applies the block diagonal matrix on the right hand side.
        Example, X @ self.

        This is the right hand side equivalent to `apply`.

        :param X: Coefficient matrix, each column is a coefficient vector.

        :return: A matrix with new coefficient vectors.
        """

        # For now do the transposes a @ b = (b.T @ a.T).T.
        #  Note there is an optimization opportunity here,
        #  but the current application of this method is only called once
        #  per FSPCA/RIR classification.
        return self.T.apply(X.T).T

    def eigvals(self):
        """
        Compute the eigenvalues of a BlkDiagMatrix.
        :return: Array of eigvals, with length equal to the fully expanded matrix diagonal.

        """
        return np.concatenate([np.linalg.eigvals(blk).flatten() for blk in self])

    def check_psd(self):
        """
        Check the positive semidefinite property of all blocks

        :return: True if all blocks have non-negative eigenvalues.
        """
        return np.alltrue(self.eigvals() > 0.0)

    def make_psd(self):
        """
        Convert all blocks to positive semidefinite

        :return: The BlkDiagMatrix instance with all blocks
            positive semidefinite
        """

        C = BlkDiagMatrix(self.partition, dtype=self.dtype)

        for i in range(self.nblocks):
            C[i] = make_psd(self[i])

        return C

    @staticmethod
    def empty(nblocks, dtype=np.float32):
        """
        Instantiate an empty BlkDiagMatrix with `nblocks`, where each
        data block is initially None with size (0,0).

        This is used for incrementally building up a BlkDiagMatrix, by
        using nblocks=0 in conjunction with `append` method or
        situations where blocks are immediately assigned in a loop,
        such as in a `copy`.

        :param nblocks: Number of diagonal matrix blocks.
        :return: BlkDiagMatrix instance where each block is None.
        """

        # Empty partition has block dims of zero until they are assigned
        partition = [(0, 0)] * nblocks

        return BlkDiagMatrix(partition, dtype=dtype)

    @staticmethod
    def zeros(blk_partition, dtype=np.float32):
        """
        Build a BlkDiagMatrix zeros matrix.

        :param blk_partition: The matrix block partition in the form of a
            K-element list storing all shapes of K diagonal matrix blocks,
            where `blk_partition[i]` corresponds to the shape (number of rows and
            columns) of the `i` diagonal matrix block.
        :param dtype: The data type to set precision of diagonal matrix block.
        :return: A BlkDiagMatrix instance consisting of `K` zero blocks.
        """

        A = BlkDiagMatrix(blk_partition, dtype=dtype)

        for i, blk_sz in enumerate(blk_partition):
            A[i] = np.zeros(blk_sz, dtype=dtype)

        return A

    @staticmethod
    def ones(blk_partition, dtype=np.float32):
        """
        Build a BlkDiagMatrix ones matrix.

        :param blk_partition: The matrix block partition in the form of a
            K-element list storing all shapes of K diagonal matrix blocks,
            where `blk_partition[i]` corresponds to the shape (number of rows and
            columns) of the `i` diagonal matrix block.
        :param dtype: The data type to set precision of diagonal matrix block.
        :return: A BlkDiagMatrix instance consisting of `K` ones blocks.
        """

        A = BlkDiagMatrix(blk_partition, dtype=dtype)

        for i, blk_sz in enumerate(blk_partition):
            A[i] = np.ones(blk_sz, dtype=dtype)

        return A

    @staticmethod
    def eye(blk_partition, dtype=np.float32):
        """
        Build a BlkDiagMatrix eye (identity) matrix

        :param blk_partition: The matrix block partition in the form of
            a K-element list storing all shapes of K diagonal matrix blocks,
            where `blk_partition[i]` corresponds to the shape (number of rows
            and columns) of the `i` diagonal matrix block.
        :param dtype: The data type of the diagonal matrix blocks.
        :return: A BlkDiagMatrix instance consisting of `K` eye (identity)
            blocks.
        """

        A = BlkDiagMatrix(blk_partition, dtype=dtype)

        for i, blk_sz in enumerate(blk_partition):
            rows, cols = blk_sz
            A[i] = np.eye(N=rows, M=cols, dtype=dtype)

        return A

    @staticmethod
    def eye_like(A, dtype=None):
        """
        Build a BlkDiagMatrix eye (identity) matrix with the partition
        structure of BlkDiagMatrix A.  Defaults to dtype of A.

        :param A: BlkDiagMatrix instance.
        :param dtype: Optional, data type of the new diagonal matrix blocks.
        :return: BlkDiagMatrix instance consisting of `K` eye (identity)
            blocks.
        """

        if dtype is None:
            dtype = A.dtype

        return BlkDiagMatrix.eye(A.partition, dtype=dtype)

    @staticmethod
    def zeros_like(A, dtype=None):
        """
        Build a BlkDiagMatrix zeros matrix with the partition
        structure of BlkDiagMatrix A.  Defaults to dtype of A.

        :param A: BlkDiagMatrix instance.
        :param dtype: Optional, data type of the new diagonal matrix blocks.
        :return: BlkDiagMatrix instance consisting of `K` zeros blocks.
        """

        if dtype is None:
            dtype = A.dtype

        return BlkDiagMatrix.zeros(A.partition, dtype=dtype)

    @staticmethod
    def from_list(blk_diag, dtype=np.float32):
        """
        Convert full from python list representation into BlkDiagMatrix.

        This is to facilitate integration with code that may not be
        using the BlkDiagMatrix class yet.

        :param blk_diag; The blk_diag representation in the form of a
            K-element list storing all shapes of K diagonal matrix blocks,
            where `blk_partition[i]` corresponds to the shape (number of rows
            and columns) of the `i` diagonal matrix block.

        :return: The BlkDiagMatrix instance.
        """

        # get the partition (just sizes)
        blk_partition = [None] * len(blk_diag)
        for i, mat in enumerate(blk_diag):
            blk_partition[i] = np.shape(mat)

        # instantiate an empty BlkDiagMatrix with that structure
        A = BlkDiagMatrix(blk_partition, dtype=dtype)

        # set the data
        for i in range(A.nblocks):
            A.data[i] = np.array(blk_diag[i], dtype=dtype)

        return A



import scipy as sp

from aspire.numeric.base_fft import FFT


class ScipyFFT(FFT):
    """
    Define a unified wrapper class for Scipy FFT functions

    To be consistent with Pyfftw, not all arguments are included.
    """

    def fft(self, x, axis=-1, workers=-1):
        return sp.fft.fft(x, axis=axis, workers=workers)

    def ifft(self, x, axis=-1, workers=-1):
        return sp.fft.ifft(x, axis=axis, workers=workers)

    def fft2(self, x, axes=(-2, -1), workers=-1):
        return sp.fft.fft2(x, axes=axes, workers=workers)

    def ifft2(self, x, axes=(-2, -1), workers=-1):
        return sp.fft.ifft2(x, axes=axes, workers=workers)

    def fftn(self, x, axes=None, workers=-1):
        return sp.fft.fftn(x, axes=axes, workers=workers)

    def ifftn(self, x, axes=None, workers=-1):
        return sp.fft.ifftn(x, axes=axes, workers=workers)

    def fftshift(self, x, axes=None):
        return sp.fft.fftshift(x, axes=axes)

    def ifftshift(self, x, axes=None):
        return sp.fft.ifftshift(x, axes=axes)



import logging

import click

from aspire.abinitio import CLSyncVoting
from aspire.commands import log_level_option
from aspire.source.relion import RelionSource
from aspire.utils.logging import setConsoleLoggingLevel

logger = logging.getLogger(__name__)


@click.command()
@click.option("--data_folder", default=None, help="Path to data folder")
@click.option(
    "--starfile_in",
    required=True,
    help="Path to input STAR file relative to data folder",
)
@click.option(
    "--starfile_out",
    required=True,
    help="Path to output STAR file relative to data folder",
)
@click.option(
    "--pixel_size", default=1, type=float, help="Pixel size of images in STAR file"
)
@click.option(
    "--max_rows",
    default=None,
    type=int,
    help="Max number of image rows to read from STAR file",
)
@click.option(
    "--n_rad",
    default=None,
    type=int,
    help="Number of points in the radial direction. If None, defaults to half the resolution of the source.",
)
@click.option(
    "--n_theta",
    default=360,
    type=int,
    help="Number of points in the theta direction",
)
@click.option(
    "--max_shift",
    default=0.15,
    type=float,
    help="Maximum range for shifts as a proportion of resolution",
)
@click.option(
    "--shift_step",
    default=1,
    type=int,
    help="Resolution for shift estimation in pixels",
)
@log_level_option
def orient3d(
    data_folder,
    starfile_in,
    starfile_out,
    pixel_size,
    max_rows,
    n_rad,
    n_theta,
    max_shift,
    shift_step,
    loglevel,
):
    """
    Input images from STAR file and estimate orientational angles
    """
    # Set desired logging option for the command line
    setConsoleLoggingLevel(loglevel)

    logger.info(f"Read in images from {starfile_in} and estimate orientational angles.")
    # Create a source object for 2D images
    source = RelionSource(
        starfile_in, data_folder, pixel_size=pixel_size, max_rows=max_rows
    )

    # Estimate rotation matrices
    logger.info("Estimate rotation matrices.")
    orient_est = CLSyncVoting(
        source, n_rad=n_rad, n_theta=n_theta, max_shift=max_shift, shift_step=shift_step
    )
    orient_est.estimate_rotations()

    # Create new source object and save Estimate rotation matrices
    logger.info("Save Estimate rotation matrices.")
    orient_est_src = orient_est.save_rotations()

    orient_est_src.save_metadata(starfile_out)
